#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML CLI.

The ZenML CLI tool is usually downloaded and installed via PyPI and a
``pip install zenml`` command. Please see the Installation & Setup
section above for more information about that process.

How to use the CLI
------------------

Our CLI behaves similarly to many other CLIs for basic features. In
order to find out which version of ZenML you are running, type:

```bash
   zenml version
```

If you ever need more information on exactly what a certain command will
do, use the ``--help`` flag attached to the end of your command string.

For example, to get a sense of all the commands available to you
while using the ``zenml`` command, type:

```bash
   zenml --help
```

If you were instead looking to know more about a specific command, you
can type something like this:

```bash
   zenml artifact-store register --help
```

This will give you information about how to register an artifact store.
(See below for more on that).

If you want to instead understand what the concept behind a group is, you 
can use the `explain` sub-command. For example, to see more details behind 
what a `artifact-store` is, you can type:

```bash
zenml artifact-store explain
```

This will give you an explanation of that concept in more detail.

Beginning a Project
-------------------

In order to start working on your project, initialize a ZenML repository
within your current directory with ZenML's own config and resource management
tools:

```bash
zenml init
```

This is all you need to begin using all the MLOps goodness that ZenML
provides!

By default, ``zenml init`` will install its own hidden ``.zen`` folder
inside the current directory from which you are running the command.
You can also pass in a directory path manually using the
``--path`` option:

```bash
zenml init --path /path/to/dir
```

If you wish to use one of [the available ZenML project templates](https://docs.zenml.io/user-guides/best-practices/project-templates)
to generate a ready-to-use project scaffold in your repository, you can do so by
passing the ``--template`` option:

```bash
zenml init --template <name_of_template>
```

Running the above command will result in input prompts being shown to you. If 
you would like to rely on default values for the ZenML project template - 
you can add ``--template-with-defaults`` to the same command, like this:

```bash
zenml init --template <name_of_template> --template-with-defaults
```

In a similar fashion, if you would like to quickly explore the capabilities
of ZenML through a notebook, you can also use:

```bash
zenml go
```

Cleaning up
-----------

If you wish to delete all data relating to your project from the
directory, use the ``zenml clean`` command. This will:

-  delete all pipelines, pipeline runs and associated metadata
-  delete all artifacts

Using Integrations
------------------

Integrations are the different pieces of a project stack that enable custom
functionality. This ranges from bigger libraries like
[`kubeflow`](https://www.kubeflow.org/) for orchestration down to smaller
visualization tools like [`facets`](https://pair-code.github.io/facets/). Our
CLI is an easy way to get started with these integrations.

To list all the integrations available to you, type:

```bash
zenml integration list
```

To see the requirements for a specific integration, use the `requirements`
command:

```bash
zenml integration requirements INTEGRATION_NAME
```

If you wish to install the integration, using the requirements listed in the
previous command, `install` allows you to do this for your local environment:

```bash
zenml integration install INTEGRATION_NAME
```

Note that if you don't specify a specific integration to be installed, the
ZenML CLI will install **all** available integrations.

If you want to install all integrations apart from one or multiple integrations,
use the following syntax, for example, which will install all integrations
except `feast` and `aws`:

```shell
zenml integration install -i feast -i aws
```

Uninstalling a specific integration is as simple as typing:

```bash
zenml integration uninstall INTEGRATION_NAME
```

For all these `zenml integration` commands, you can pass the `--uv` flag and we
will use `uv` as the package manager instead of `pip`. This will resolve and
install much faster than with `pip`, but note that it requires `uv` to be
installed on your machine. This is an experimental feature and may not work on
all systems. In particular, note that installing onto machines with GPU
acceleration may not work as expected.

If you would like to export the requirements of all ZenML integrations, you can
use the command:

```bash
zenml integration export-requirements
```

Here, you can also select a list of integrations and write the result into and
output file:

```bash
zenml integration export-requirements gcp kubeflow -o OUTPUT_FILE
```

Filtering when listing
----------------------

Certain CLI `list` commands allow you to filter their output. For example, all
stack components allow you to pass custom parameters to the `list` command that
will filter the output. To learn more about the available filters, a good quick
reference is to use the `--help` command, as in the following example:

```shell
zenml orchestrator list --help
```

You will see a list of all the available filters for the `list` command along
with examples of how to use them.

The `--sort_by` option allows you to sort the output by a specific field and
takes an `asc` or `desc` argument to specify the order. For example, to sort the
output of the `list` command by the `name` field in ascending order, you would
type:

```shell
zenml orchestrator list --sort_by "asc:name"
```

For fields marked as being of type `TEXT` or `UUID`, you can use the `contains`,
`startswith` and `endswith` keywords along with their particular identifier. For
example, for the orchestrator `list` command, you can use the following filter
to find all orchestrators that contain the string `sagemaker` in their name:

```shell
zenml orchestrator list --name "contains:sagemaker"
```

For fields marked as being of type `BOOL`, you can use the 'True' or 'False'
values to filter the output.

Finally, for fields marked as being of type `DATETIME`, you can pass in datetime
values in the `%Y-%m-%d %H:%M:%S` format. These can be combined with the `gte`,
`lte`, `gt` and `lt` keywords (greater than or equal, less than or equal,
greater than and less than respectively) to specify the range of the filter. For
example, if I wanted to find all orchestrators that were created after the 1st
of January 2021, I would type:

```shell
zenml orchestrator list --created "gt:2021-01-01 00:00:00"
```

This syntax can also be combined to create more complex filters using the `or`
and `and` keywords.

Artifact Stores
---------------

In ZenML, [the artifact store](https://docs.zenml.io/stacks/stack-components/artifact-stores)
is where all the inputs and outputs of your pipeline steps are stored. By
default, ZenML initializes your repository with an artifact store with
everything kept on your local machine. You can get a better understanding
about the concept of artifact stores by executing:

```bash
zenml artifact-store explain
```

If you wish to register a new artifact store, do so with the ``register``
command:

```bash
zenml artifact-store register ARTIFACT_STORE_NAME --flavor=ARTIFACT_STORE_FLAVOR [--OPTIONS]
```

You can also add any labels to your stack component using the `--label` or `-l` flag:

```bash
zenml artifact-store register ARTIFACT_STORE_NAME --flavor=ARTIFACT_STORE_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new artifact store,
you have to choose a flavor. To see the full list of available artifact
store flavors, you can use the command:

```bash
zenml artifact-store flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml artifact-store flavor describe FLAVOR_NAME
```

If you wish to list the artifact stores that have already been registered
within your ZenML:

```bash
zenml artifact-store list
```

If you want the name of the artifact store in the active stack, you can
also use the `get` command:

```bash
zenml artifact-store get
```

For details about a particular artifact store, use the `describe` command.
By default, (without a specific artifact store name passed in) it will describe
the active or currently used artifact store:

```bash
zenml artifact-store describe ARTIFACT_STORE_NAME
```

If you wish to update/rename an artifact store, you can use the following
commands respectively:

```bash
zenml artifact-store update ARTIFACT_STORE_NAME --property_to_update=new_value
zenml artifact-store rename ARTIFACT_STORE_OLD_NAME ARTIFACT_STORE_NEW_NAME
```

If you wish to delete a particular artifact store, pass the name of the
artifact store into the CLI with the following command:

```bash
zenml artifact-store delete ARTIFACT_STORE_NAME
```

If you would like to connect/disconnect your artifact store to/from a service
connector, you can use the following commands:

```bash
zenml artifact-store connect ARTIFACT_STORE_NAME -c CONNECTOR_NAME
zenml artifact-store disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
artifact stores. In order to get a full list of available functions, use
the command:

```bash
zenml artifact-store --help
```

Orchestrators
-------------

An [orchestrator](https://docs.zenml.io/stacks/stack-components/orchestrators)
is a special kind of backend that manages the running of each step of the
pipeline. Orchestrators administer the actual pipeline runs. By default,
ZenML initializes your repository with an orchestrator that runs everything
on your local machine. In order to get a more detailed explanation, you can use
the command:

```bash
zenml orchestrator explain
```

If you wish to register a new orchestrator, do so with the ``register``
command:

```bash
zenml orchestrator register ORCHESTRATOR_NAME --flavor=ORCHESTRATOR_FLAVOR [--ORCHESTRATOR_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml orchestrator register ORCHESTRATOR_NAME --flavor=ORCHESTRATOR_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new orchestrator,
you have to choose a flavor. To see the full list of available orchestrator
flavors, you can use the command:

```bash
zenml orchestrator flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml orchestrator flavor describe FLAVOR_NAME
```

If you wish to list the orchestrators that have already been registered, type:

```bash
zenml orchestrator list
```

If you want the name of the orchestrator in the active stack, you can
also use the `get` command:

```bash
zenml orchestrator get
```

For details about a particular orchestrator, use the `describe` command.
By default, (without a specific orchestrator name passed in) it will describe
the active or currently used orchestrator:

```bash
zenml orchestrator describe [ORCHESTRATOR_NAME]
```

If you wish to update/rename an orchestrator, you can use the following
commands respectively:

```bash
zenml orchestrator update ORCHESTRATOR_NAME --property_to_update=new_value
zenml orchestrator rename ORCHESTRATOR_OLD_NAME ORCHESTRATOR_NEW_NAME
```

If you wish to delete a particular orchestrator, pass the name of the
orchestrator into the CLI with the following command:

```bash
zenml orchestrator delete ORCHESTRATOR_NAME
```

If you would like to connect/disconnect your orchestrator to/from a service
connector, you can use the following commands:

```bash
zenml orchestrator connect ORCHESTRATOR_NAME -c CONNECTOR_NAME
zenml orchestrator disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
orchestrators. In order to get a full list of available functions, use
the command:

```bash
zenml orchestrators --help
```

Container Registries
--------------------

[The container registry](https://docs.zenml.io/stacks/stack-components/container-registries)
is where all the images that are used by a container-based orchestrator are
stored. To get a better understanding regarding container registries, use
the command:

```bash
zenml container-registry explain
```

By default, a default ZenML local stack will not register a container registry.
If you wish to register a new container registry, do so with the `register`
command:

```bash
zenml container-registry register REGISTRY_NAME --flavor=REGISTRY_FLAVOR [--REGISTRY_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml container-registry register REGISTRY_NAME --flavor=REGISTRY_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new container
registry, you have to choose a flavor. To see the full list of available
container registry flavors, you can use the command:

```bash
zenml container-registry flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml container-registry flavor describe FLAVOR_NAME
```

To list all container registries available and registered for use, use the
`list` command:

```bash
zenml container-registry list
```

If you want the name of the container registry in the active stack, you can
also use the `get` command:

```bash
zenml container-registry get
```

For details about a particular container registry, use the `describe` command.
By default, (without a specific registry name passed in) it will describe the
active or currently used container registry:

```bash
zenml container-registry describe [CONTAINER_REGISTRY_NAME]
```

If you wish to update/rename a container registry, you can use the following
commands respectively:

```bash
zenml container-registry update CONTAINER_REGISTRY_NAME --property_to_update=new_value
zenml container-registry rename CONTAINER_REGISTRY_OLD_NAME CONTAINER_REGISTRY_NEW_NAME
```

To delete a container registry (and all of its contents), use the `delete`
command:

```bash
zenml container-registry delete REGISTRY_NAME
```

If you would like to connect/disconnect your container registry to/from a
service connector, you can use the following commands:

```bash
zenml container-registry connect CONTAINER_REGISTRY_NAME -c CONNECTOR_NAME
zenml container-registry disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
container registries. In order to get a full list of available functions,
use the command:

```bash
zenml container-registry --help
```

Data Validators
---------------

In ZenML, [data validators](https://docs.zenml.io/stacks/stack-components/data-validators)
help you profile and validate your data.

By default, a default ZenML local stack will not register a data validator. If
you wish to register a new data validator, do so with the `register` command:

```bash
zenml data-validator register DATA_VALIDATOR_NAME --flavor DATA_VALIDATOR_FLAVOR [--DATA_VALIDATOR_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml data-validator register DATA_VALIDATOR_NAME --flavor DATA_VALIDATOR_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new data validator,
you have to choose a flavor. To see the full list of available data validator
flavors, you can use the command:

```bash
zenml data-validator flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml data-validator flavor describe FLAVOR_NAME
```

To list all data validators available and registered for use, use the `list`
command:

```bash
zenml data-validator list
```

If you want the name of the data validator in the active stack, use the `get`
command:

```bash
zenml data-validator get
```

For details about a particular data validator, use the `describe` command.
By default, (without a specific data validator name passed in) it will describe
the active or currently-used data validator:

```bash
zenml data-validator describe [DATA_VALIDATOR_NAME]
```

If you wish to update/rename a data validator, you can use the following
commands respectively:

```bash
zenml data-validator update DATA_VALIDATOR_NAME --property_to_update=new_value
zenml data-validator rename DATA_VALIDATOR_OLD_NAME DATA_VALIDATOR_NEW_NAME
```

To delete a data validator (and all of its contents), use the `delete` command:

```bash
zenml data-validator delete DATA_VALIDATOR_NAME
```

If you would like to connect/disconnect your data validator to/from a service
connector, you can use the following commands:

```bash
zenml data-validator connect DATA_VALIDATOR_NAME -c CONNECTOR_NAME
zenml data-validator disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
data validators. In order to get a full list of available functions, use the
command:

```bash
zenml data-validator --help
```

Experiment Trackers
-------------------

[Experiment trackers](https://docs.zenml.io/stacks/stack-components/experiment-trackers)
 let you track your ML experiments by logging the parameters
and allow you to compare between different runs. To get a better
understanding regarding experiment trackers, use the command:

```bash
zenml experiment-tracker explain
```

By default, a default ZenML local stack will not register an experiment tracker.
If you want to use an experiment tracker in one of your stacks, you need to
first register it:

```bash
zenml experiment-tracker register EXPERIMENT_TRACKER_NAME \
    --flavor=EXPERIMENT_TRACKER_FLAVOR [--EXPERIMENT_TRACKER_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml experiment-tracker register EXPERIMENT_TRACKER_NAME \
      --flavor=EXPERIMENT_TRACKER_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new experiment
tracker, you have to choose a flavor. To see the full list of available
experiment tracker flavors, you can use the command:

```bash
zenml experiment-tracker flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml experiment-tracker flavor describe FLAVOR_NAME
```

To list all experiment trackers available and registered for use, use the
`list` command:

```bash
zenml experiment-tracker list
```

If you want the name of the experiment tracker in the active stack, use the
`get` command:

```bash
zenml experiment-tracker get
```

For details about a particular experiment tracker, use the `describe` command.
By default, (without a specific experiment tracker name passed in) it will
describe the active or currently-used experiment tracker:

```bash
zenml experiment-tracker describe [EXPERIMENT_TRACKER_NAME]
```

If you wish to update/rename an experiment tracker, you can use the following
commands respectively:

```bash
zenml experiment-tracker update EXPERIMENT_TRACKER_NAME --property_to_update=new_value
zenml experiment-tracker rename EXPERIMENT_TRACKER_OLD_NAME EXPERIMENT_TRACKER_NEW_NAME
```

To delete an experiment tracker, use the `delete` command:

```bash
zenml experiment-tracker delete EXPERIMENT_TRACKER_NAME
```

If you would like to connect/disconnect your experiment tracker to/from a
service connector, you can use the following commands:

```bash
zenml experiment-tracker connect EXPERIMENT_TRACKER_NAME -c CONNECTOR_NAME
zenml experiment-tracker disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
experiment trackers. In order to get a full list of available functions,
use the command:

```bash
zenml experiment-tracker --help
```

Model Deployers
---------------

[Model deployers](https://docs.zenml.io/stacks/stack-components/model-deployers)
are stack components responsible for online model serving. They are responsible
for deploying models to a remote server. Model deployers also act as a registry
for models that are served with ZenML. To get a better understanding regarding
model deployers, use the command:

```bash
zenml model-deployer explain
```

By default, a default ZenML local stack will not register a model deployer. If
you wish to register a new model deployer, do so with the `register` command:

```bash
zenml model-deployer register MODEL_DEPLOYER_NAME --flavor MODEL_DEPLOYER_FLAVOR [--MODEL_DEPLOYER_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml model-deployer register MODEL_DEPLOYER_NAME --flavor MODEL_DEPLOYER_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new model deployer,
you have to choose a flavor. To see the full list of available model deployer
flavors, you can use the command:

```bash
zenml model-deployer flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml model-deployer flavor describe FLAVOR_NAME
```

To list all model deployers available and registered for use, use the
`list` command:

```bash
zenml model-deployer list
```

If you want the name of the model deployer in the active stack, use the `get`
command:

```bash
zenml model-deployer get
```

For details about a particular model deployer, use the `describe` command.
By default, (without a specific operator name passed in) it will describe the
active or currently used model deployer:

```bash
zenml model-deployer describe [MODEL_DEPLOYER_NAME]
```

If you wish to update/rename a model deployer, you can use the following
commands respectively:

```bash
zenml model-deployer update MODEL_DEPLOYER_NAME --property_to_update=new_value
zenml model-deployer rename MODEL_DEPLOYER_OLD_NAME MODEL_DEPLOYER_NEW_NAME
```

To delete a model deployer (and all of its contents), use the `delete`
command:

```bash
zenml model-deployer delete MODEL_DEPLOYER_NAME
```

If you would like to connect/disconnect your model deployer to/from a
service connector, you can use the following commands:

```bash
zenml model-deployer connect MODEL_DEPLOYER_NAME -c CONNECTOR_NAME
zenml model-deployer disconnect
```

Moreover, ZenML features a set of CLI commands specific to the model deployer
interface. If you want to simply see what models have been deployed within
your stack, run the following command:

```bash
zenml model-deployer models list
```

This should give you a list of served models containing their `uuid`, the name
of the pipeline that produced them including the run id and the step name as
well as the status. This information should help you identify the different
models.

If you want further information about a specific model, simply copy the
UUID and the following command.

```bash
zenml model-deployer models describe <UUID>
```

If you are only interested in the prediction url of the specific model you can
also run:

```bash
zenml model-deployer models get-url <UUID>
```

Finally, you will also be able to start/stop the services using the following
 two commands:

```bash
zenml model-deployer models start <UUID>
zenml model-deployer models stop <UUID>
```

If you want to completely remove a served model you can also irreversibly delete
 it using:

```bash
zenml model-deployer models delete <UUID>
```

The ZenML CLI provides a few more utility functions for you to manage your
model deployers. In order to get a full list of available functions,
use the command:

```bash
zenml model-deployer --help
```

Step Operators
--------------

[Step operators](https://docs.zenml.io/stacks/stack-components/step-operators)
allow you to run individual steps in a custom environment different from the
default one used by your active orchestrator. One example use-case is to run a
training step of your pipeline in an environment with GPUs available. To get
a better understanding regarding step operators, use the command:

```bash
zenml step-operator explain
```

By default, a default ZenML local stack will not register a step operator. If
you wish to register a new step operator, do so with the `register` command:

```bash
zenml step-operator register STEP_OPERATOR_NAME --flavor STEP_OPERATOR_FLAVOR [--STEP_OPERATOR_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml step-operator register STEP_OPERATOR_NAME --flavor STEP_OPERATOR_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new step operator,
you have to choose a flavor. To see the full list of available step operator
flavors, you can use the command:

```bash
zenml step-operator flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml step-operator flavor describe FLAVOR_NAME
```

To list all step operators available and registered for use, use the
`list` command:

```bash
zenml step-operator list
```

If you want the name of the step operator in the active stack, use the `get`
command:

```bash
zenml step-operator get
```

For details about a particular step operator, use the `describe` command.
By default, (without a specific operator name passed in) it will describe the
active or currently used step operator:

```bash
zenml step-operator describe [STEP_OPERATOR_NAME]
```

If you wish to update/rename a step operator, you can use the following commands
respectively:

```bash
zenml step-operator update STEP_OPERATOR_NAME --property_to_update=new_value
zenml step-operator rename STEP_OPERATOR_OLD_NAME STEP_OPERATOR_NEW_NAME
```

To delete a step operator (and all of its contents), use the `delete`
command:

```bash
zenml step-operator delete STEP_OPERATOR_NAME
```

If you would like to connect/disconnect your step operator to/from a
service connector, you can use the following commands:

```bash
zenml step-operator connect STEP_OPERATOR_NAME -c CONNECTOR_NAME
zenml step-operator disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
step operators. In order to get a full list of available functions,
use the command:

```bash
zenml step-operator --help
```

Alerters
--------

In ZenML, [alerters](https://docs.zenml.io/stacks/stack-components/alerters)
allow you to send alerts from within your pipeline.

By default, a default ZenML local stack will not register an alerter. If
you wish to register a new alerter, do so with the `register` command:

```bash
zenml alerter register ALERTER_NAME --flavor ALERTER_FLAVOR [--ALERTER_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml alerter register ALERTER_NAME --flavor ALERTER_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new alerter,
you have to choose a flavor. To see the full list of available alerter
flavors, you can use the command:

```bash
zenml alerter flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml alerter flavor describe FLAVOR_NAME
```

To list all alerters available and registered for use, use the `list` command:

```bash
zenml alerter list
```

If you want the name of the alerter in the active stack, use the `get`
command:

```bash
zenml alerter get
```

For details about a particular alerter, use the `describe` command.
By default, (without a specific alerter name passed in) it will describe
the active or currently used alerter:

```bash
zenml alerter describe [ALERTER_NAME]
```

If you wish to update/rename an alerter, you can use the following commands
respectively:

```bash
zenml alerter update ALERTER_NAME --property_to_update=new_value
zenml alerter rename ALERTER_OLD_NAME ALERTER_NEW_NAME
```

To delete an alerter (and all of its contents), use the `delete` command:

```bash
zenml alerter delete ALERTER_NAME
```

If you would like to connect/disconnect your alerter to/from a service
connector, you can use the following commands:

```bash
zenml alerter connect ALERTER_NAME -c CONNECTOR_NAME
zenml alerter disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
alerters. In order to get a full list of available functions, use the command:

```bash
zenml alerter --help
```

Feature Stores
--------------

[Feature stores](https://docs.zenml.io/stacks/stack-components/feature-stores)
allow data teams to serve data via an offline store and an online low-latency
store where data is kept in sync between the two. To get a better understanding
regarding feature stores, use the command:

```bash
zenml feature-store explain
```

By default, a default ZenML local stack will not register a feature store. If
you wish to register a new feature store, do so with the `register` command:

```bash
zenml feature-store register FEATURE_STORE_NAME --flavor FEATURE_STORE_FLAVOR [--FEATURE_STORE_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml feature-store register FEATURE_STORE_NAME --flavor FEATURE_STORE_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new feature store,
you have to choose a flavor. To see the full list of available feature store
flavors, you can use the command:

```bash
zenml feature-store flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

Note: Currently, ZenML only supports connecting to a Redis-backed Feast feature
store as a stack component integration.

```bash
zenml feature-store flavor describe FLAVOR_NAME
```

To list all feature stores available and registered for use, use the
`list` command:

```bash
zenml feature-store list
```

If you want the name of the feature store in the active stack, use the `get`
command:

```bash
zenml feature-store get
```

For details about a particular feature store, use the `describe` command.
By default, (without a specific feature store name passed in) it will describe
the active or currently-used feature store:

```bash
zenml feature-store describe [FEATURE_STORE_NAME]
```

If you wish to update/rename a feature store, you can use the following commands
respectively:

```bash
zenml feature-store update FEATURE_STORE_NAME --property_to_update=new_value
zenml feature-store rename FEATURE_STORE_OLD_NAME FEATURE_STORE_NEW_NAME
```

To delete a feature store (and all of its contents), use the `delete`
command:

```bash
zenml feature-store delete FEATURE_STORE_NAME
```

If you would like to connect/disconnect your feature store to/from a
service connector, you can use the following commands:

```bash
zenml feature-store connect FEATURE_STORE_NAME -c CONNECTOR_NAME
zenml feature-store disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
feature stores. In order to get a full list of available functions,
use the command:

```bash
zenml feature-store --help
```

Annotators
----------

[Annotators](https://docs.zenml.io/stacks/stack-components/annotators)
enable the use of data annotation as part of your ZenML stack and pipelines.

By default, a default ZenML local stack will not register an annotator. If
you wish to register a new annotator, do so with the `register` command:

```bash
zenml annotator register ANNOTATOR_NAME --flavor ANNOTATOR_FLAVOR [--ANNOTATOR_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml annotator register ANNOTATOR_NAME --flavor ANNOTATOR_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new annotator,
you have to choose a flavor. To see the full list of available annotator
flavors, you can use the command:

```bash
zenml annotator flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml annotator flavor describe FLAVOR_NAME
```

To list all annotator available and registered for use, use the `list` command:

```bash
zenml annotator list
```

If you want the name of the annotator in the active stack, use the `get`
command:

```bash
zenml annotator get
```

For details about a particular annotator, use the `describe` command.
By default, (without a specific annotator name passed in) it will describe
the active or currently used annotator:

```bash
zenml annotator describe [ANNOTATOR_NAME]
```

If you wish to update/rename an annotator, you can use the following commands
respectively:

```bash
zenml annotator update ANNOTATOR_NAME --property_to_update=new_value
zenml annotator rename ANNOTATOR_OLD_NAME ANNOTATOR_NEW_NAME
```

To delete an annotator (and all of its contents), use the `delete` command:

```bash
zenml annotator delete ANNOTATOR_NAME
```

If you would like to connect/disconnect your annotator to/from a service
connector, you can use the following commands:

```bash
zenml annotator connect ANNOTATOR_NAME -c CONNECTOR_NAME
zenml annotator disconnect
```

Finally, you can use the `dataset` command to interact with your annotation
datasets:

```bash
zenml annotator dataset --help
```

The ZenML CLI provides a few more utility functions for you to manage your
annotator. In order to get a full list of available functions, use the command:

```bash
zenml annotator --help
```

Image Builders
--------------

In ZenML, [image builders](https://docs.zenml.io/stacks/stack-components/image-builders)
allow you to build container images such
that your machine-learning pipelines and steps can be executed in remote
environments.

By default, a default ZenML local stack will not register an image builder. If
you wish to register a new image builder, do so with the `register` command:

```bash
zenml image-builder register IMAGE_BUILDER_NAME --flavor IMAGE_BUILDER_FLAVOR [--IMAGE_BUILDER_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml image-builder register IMAGE_BUILDER_NAME --flavor IMAGE_BUILDER_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new image builder,
you have to choose a flavor. To see the full list of available image builder
flavors, you can use the command:

```bash
zenml image-builder flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml image-builder flavor describe FLAVOR_NAME
```

To list all image builders available and registered for use, use the `list`
command:

```bash
zenml image-builder list
```

If you want the name of the image builder in the active stack, use the `get`
command:

```bash
zenml image-builder get
```

For details about a particular image builder, use the `describe` command.
By default, (without a specific image builder name passed in) it will describe
the active or currently used image builder:

```bash
zenml image-builder describe [IMAGE_BUILDER_NAME]
```

If you wish to update/rename an image builder, you can use the following
commands respectively:

```bash
zenml image-builder update IMAGE_BUILDER_NAME --property_to_update=new_value
zenml image-builder rename IMAGE_BUILDER_OLD_NAME IMAGE_BUILDER_NEW_NAME
```

To delete a image builder (and all of its contents), use the `delete` command:

```bash
zenml image-builder delete IMAGE_BUILDER_NAME
```

If you would like to connect/disconnect your image builder to/from a service
connector, you can use the following commands:

```bash
zenml image-builder connect IMAGE_BUILDER_NAME -c CONNECTOR_NAME
zenml image-builder disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
image builders. In order to get a full list of available functions, use the
command:

```bash
zenml image-builder --help
```

Model Registries
----------------

[Model registries](https://docs.zenml.io/stacks/stack-components/model-registries)
are centralized repositories that facilitate the collaboration and management
of machine learning models. To get a better understanding regarding model
registries as a concept, use the command:

```bash
zenml model-registry explain
```

By default, a default ZenML local stack will not register a model registry. If
you wish to register a new model registry, do so with the `register` command:

```bash
zenml model-registry register MODEL_REGISTRY_NAME --flavor MODEL_REGISTRY_FLAVOR [--MODEL_REGISTRY_OPTIONS]
```

You can also add any label to your stack component using the `--label` or `-l` flag:

```bash
zenml model-registry register MODEL_REGISTRY_NAME --flavor MODEL_REGISTRY_FLAVOR -l key1=value1 -l key2=value2
```

As you can see from the command above, when you register a new model registry,
you have to choose a flavor. To see the full list of available model registry
flavors, you can use the command:

```bash
zenml model-registry flavor list
```

This list will show you which integration these flavors belong to and which
service connectors they are adaptable with. If you would like to get additional
information regarding a specific flavor, you can utilize the command:

```bash
zenml model-registry flavor describe FLAVOR_NAME
```

To list all model registries available and registered for use, use the
`list` command:

```bash
zenml model-registry list
```

If you want the name of the model registry in the active stack, use the `get`
command:

```bash
zenml model-registry get
```

For details about a particular model registry, use the `describe` command.
By default, (without a specific operator name passed in) it will describe the
active or currently used model registry:

```bash
zenml model-registry describe [MODEL_REGISTRY_NAME]
```

If you wish to update/rename a model registry, you can use the following commands
respectively:

```bash
zenml model-registry update MODEL_REGISTRY_NAME --property_to_update=new_value
zenml model-registry rename MODEL_REGISTRY_OLD_NAME MODEL_REGISTRY_NEW_NAME
```

To delete a model registry (and all of its contents), use the `delete`
command:

```bash
zenml model-registry delete MODEL_REGISTRY_NAME
```

If you would like to connect/disconnect your model registry to/from a
service connector, you can use the following commands:

```bash
zenml model-registry connect MODEL_REGISTRY_NAME -c CONNECTOR_NAME
zenml model-registry disconnect
```

The ZenML CLI provides a few more utility functions for you to manage your
model registries. In order to get a full list of available functions,
use the command:

```bash
zenml model-registry --help
```

Managing your Stacks
--------------------

[The stack](https://docs.zenml.io/user-guides/production-guide/understand-stacks)
is a grouping of your artifact store, your orchestrator, and other
optional MLOps tools like experiment trackers or model deployers.
With the ZenML tool, switching from a local stack to a distributed cloud
environment can be accomplished with just a few CLI commands.

To register a new stack, you must already have registered the individual
components of the stack using the commands listed above.

Use the ``zenml stack register`` command to register your stack. It
takes four arguments as in the following example:

```bash
zenml stack register STACK_NAME \
       -a ARTIFACT_STORE_NAME \
       -o ORCHESTRATOR_NAME
```

Each corresponding argument should be the name, id or even the first few letters
 of the id that uniquely identify the artifact store or orchestrator.

To create a new stack using the new service connector with a set of minimal components, 
use the following command:

```bash
zenml stack register STACK_NAME \
       -p CLOUD_PROVIDER
```

To create a new stack using the existing service connector with a set of minimal components, 
use the following command:

```bash
zenml stack register STACK_NAME \
       -sc SERVICE_CONNECTOR_NAME
```

To create a new stack using the existing service connector with existing components (
important, that the components are already registered in the service connector), use the 
following command:

```bash
zenml stack register STACK_NAME \
       -sc SERVICE_CONNECTOR_NAME \
       -a ARTIFACT_STORE_NAME \
       -o ORCHESTRATOR_NAME \
       ...
```

If you want to immediately set this newly created stack as your active stack,
simply pass along the `--set` flag.

```bash
zenml stack register STACK_NAME ... --set
```

To list the stacks that you have registered, type:

```bash
zenml stack list
```
To delete a stack that you have previously registered, type:

```bash
zenml stack delete STACK_NAME
```
By default, ZenML uses a local stack whereby all pipelines run on your
local computer. If you wish to set a different stack as the current
active stack to be used when running your pipeline, type:

```bash
zenml stack set STACK_NAME
```
This changes a configuration property within your local environment.

To see which stack is currently set as the default active stack, type:

```bash
zenml stack get
```

If you want to copy a stack, run the following command:
```shell
zenml stack copy SOURCE_STACK_NAME TARGET_STACK_NAME
```

If you wish to transfer one of your stacks to another machine, you can do so
by exporting the stack configuration and then importing it again.

To export a stack to YAML, run the following command:

```bash
zenml stack export STACK_NAME FILENAME.yaml
```

This will create a FILENAME.yaml containing the config of your stack and all
of its components, which you can then import again like this:

```bash
zenml stack import STACK_NAME -f FILENAME.yaml
```

If you wish to update a stack that you have already registered, first make sure
you have registered whatever components you want to use, then use the following
command:

```bash
# assuming that you have already registered a new orchestrator
# with NEW_ORCHESTRATOR_NAME
zenml stack update STACK_NAME -o NEW_ORCHESTRATOR_NAME
```

You can update one or many stack components at the same time out of the ones
that ZenML supports. To see the full list of options for updating a stack, use
the following command:

```bash
zenml stack update --help
```

To remove a stack component from a stack, use the following command:

```shell
# assuming you want to remove the image builder and the feature-store
# from your stack
zenml stack remove-component -i -f
```

If you wish to rename your stack, use the following command:

```shell
zenml stack rename STACK_NAME NEW_STACK_NAME
```

If you would like to export the requirements of your stack, you can
use the command:

```bash
zenml stack export-requirements <STACK_NAME>
```

If you want to copy a stack component, run the following command:
```bash
zenml STACK_COMPONENT copy SOURCE_COMPONENT_NAME TARGET_COMPONENT_NAME
```

If you wish to update a specific stack component, use the following command,
switching out "STACK_COMPONENT" for the component you wish to update (i.e.
'orchestrator' or 'artifact-store' etc.):

```shell
zenml STACK_COMPONENT update --some_property=NEW_VALUE
```

Note that you are not permitted to update the stack name or UUID in this way. To
change the name of your stack component, use the following command:

```shell
zenml STACK_COMPONENT rename STACK_COMPONENT_NAME NEW_STACK_COMPONENT_NAME
```

If you wish to remove an attribute (or multiple attributes) from a stack
component, use the following command:

```shell
zenml STACK_COMPONENT remove-attribute STACK_COMPONENT_NAME ATTRIBUTE_NAME [OTHER_ATTRIBUTE_NAME]
```

Note that you can only remove optional attributes.

If you want to register secrets for all secret references in a stack, use the
following command:

```shell
zenml stack register-secrets [<STACK_NAME>]
```

If you want to connect a service connector to a stack's components, you can use
the `connect` command:

```shell
zenml stack connect STACK_NAME -c CONNECTOR_NAME
```

Note that this only connects the service connector to the current components
of the stack and not to the stack itself, which means that you need to rerun
the command after adding new components to the stack.

The ZenML CLI provides a few more utility functions for you to manage your
stacks. In order to get a full list of available functions, use the command:

```bash
zenml stack --help
```

Managing your Models
--------------------

ZenML provides several CLI commands to help you administer your models and
their versions as part of [the Model Control Plane](https://docs.zenml.io/user-guides/starter-guide/track-ml-models).

To register a new model, you can use the following CLI command:

```bash
zenml model register --name <NAME> [--MODEL_OPTIONS]
```

To list all registered models, use:

```bash
zenml model list [MODEL_FILTER_OPTIONS]
```

To update a model, use:

```bash
zenml model update <MODEL_NAME_OR_ID> [--MODEL_OPTIONS]
```

If you would like to add or remove tags from the model, use:

```bash
zenml model update <MODEL_NAME_OR_ID> --tag <TAG> --tag <TAG> ..
   --remove-tag <TAG> --remove-tag <TAG> ..
```

To delete a model, use:

```bash
zenml model delete <MODEL_NAME_OR_ID>
```

The CLI interface for models also helps to navigate through artifacts linked to a specific model versions.

```bash
zenml model data_artifacts <MODEL_NAME_OR_ID> [-v <VERSION>]
zenml model deployment_artifacts <MODEL_NAME_OR_ID> [-v <VERSION>]
zenml model model_artifacts <MODEL_NAME_OR_ID> [-v <VERSION>]
```

You can also navigate the pipeline runs linked to a specific model versions:

```bash
zenml model runs <MODEL_NAME_OR_ID> [-v <VERSION>]
```

To list the model versions of a specific model, use:

```bash
zenml model version list [--model-name <MODEL_NAME> --name <MODEL_VERSION_NAME> OTHER_OPTIONS]
```

To delete a model version, use:

```bash
zenml model version delete <MODEL_NAME_OR_ID> <VERSION>
```

To update a model version, use:

```bash
zenml model version update <MODEL_NAME_OR_ID> <VERSION> [--MODEL_VERSION_OPTIONS]
```

These are some of the more common uses of model version updates:

- stage (i.e. promotion)

```bash
zenml model version update <MODEL_NAME_OR_ID> <VERSION> --stage <STAGE>
```

- tags

```bash
zenml model version update <MODEL_NAME_OR_ID> <VERSION> --tag <TAG> --tag <TAG> ..
   --remove-tag <TAG> --remove-tag <TAG> ..
```

Managing your Pipelines & Artifacts
-----------------------------------

ZenML provides several CLI commands to help you [administer your pipelines and
pipeline runs](https://docs.zenml.io/user-guides/starter-guide/manage-artifacts).

To explicitly register a pipeline you need to point to a pipeline instance
in your Python code. Let's say you have a Python file called `run.py` and
it contains the following code:

```python
from zenml import pipeline

@pipeline
def my_pipeline(...):
   # Connect your pipeline steps here
   pass
```

You can register your pipeline like this:
```bash
zenml pipeline register run.my_pipeline
```

To list all registered pipelines, use:

```bash
zenml pipeline list
```

To delete a pipeline, run:

```bash
zenml pipeline delete <PIPELINE_NAME>
```

This will delete the pipeline and change all corresponding
pipeline runs to become unlisted (not linked to any pipeline).


To list all pipeline runs that you have executed, use:

```bash
zenml pipeline runs list
```

To delete a pipeline run, use:

```bash
zenml pipeline runs delete <PIPELINE_RUN_NAME_OR_ID>
```

To refresh the status of a pipeline run, you can use the `refresh` command (
only supported for pipelines executed on Vertex, Sagemaker or AzureML).

```bash
zenml pipeline runs refresh <PIPELINE_RUN_NAME_OR_ID>
```

If you run any of your pipelines with `pipeline.run(schedule=...)`, ZenML keeps
track of the schedule and you can list all schedules via:

```bash
zenml pipeline schedule list
```

To delete a schedule, use:

```bash
zenml pipeline schedule delete <SCHEDULE_NAME_OR_ID>
```

Note, however, that this will only delete the reference saved in ZenML and does
NOT stop/delete the schedule in the respective orchestrator. This still needs to
be done manually. For example, using the Airflow orchestrator you would have
to open the web UI to manually click to stop the schedule from executing.

Each pipeline run automatically saves its artifacts in the artifact store. To
list all artifacts that have been saved, use:

```bash
zenml artifact list
```

Each artifact has one or several versions. To list artifact versions, use:

```bash
zenml artifact versions list
```

If you would like to rename an artifact or adjust the tags of an artifact or
artifact version, use the corresponding `update` command:

```bash
zenml artifact update <NAME> -n <NEW_NAME>
zenml artifact update <NAME> -t <TAG1> -t <TAG2> -r <TAG_TO_REMOVE>
zenml artifact version update <NAME> -v <VERSION> -t <TAG1> -t <TAG2> -r <TAG_TO_REMOVE>
```

The metadata of artifacts or artifact versions stored by ZenML can only be
deleted once they are no longer used by any pipeline runs. I.e., an artifact
version can only be deleted if the run that produced it and all runs that used
it as an input have been deleted. Similarly, an artifact can only be deleted if
all its versions can be deleted.

To delete all artifacts and artifact versions that are no longer linked to any
pipeline runs, use:

```bash
zenml artifact prune
```

You might find that some artifacts throw errors when you try to prune them,
likely because they were stored locally and no longer exist. If you wish to
continue pruning and to ignore these errors, please add the `--ignore-errors`
flag. Warning messages will still be output to the terminal during this
process.

Each pipeline run that requires Docker images also stores a build which
contains the image names used for this run. To list all builds, use:

```bash
zenml pipeline builds list
```

To delete a specific build, use:

```bash
zenml pipeline builds delete <BUILD_ID>
```

Managing the local ZenML Dashboard
----------------------------------

The ZenML dashboard is a web-based UI that allows you to visualize and navigate
the stack configurations, pipelines and pipeline runs tracked by ZenML among
other things. You can start the ZenML dashboard locally by running the following
command:

```bash
zenml login --local
```

This will start the dashboard on your local machine where you can access it at
the URL printed to the console.

If you have closed the dashboard in your browser and want to open it again,
you can run:

```bash
zenml show
```

If you want to stop the dashboard, simply run:

```bash
zenml logout --local
```

The `zenml login --local` command has a few additional options that you can use
to customize how the ZenML dashboard is running.

By default, the dashboard is started as a background process. On some operating
systems, this capability is not available. In this case, you can use the
`--blocking` flag to start the dashboard in the foreground:

```bash
zenml login --local --blocking
```

This will block the terminal until you stop the dashboard with CTRL-C.

Another option you can use, if you have Docker installed on your machine, is to
run the dashboard in a Docker container. This is useful if you don't want to
install all the Zenml server dependencies on your machine. To do so, simply run:

```bash
zenml login --local --docker
```

The TCP port and the host address that the dashboard uses to listen for
connections can also be customized. Using an IP address that is not the default
`localhost` or 127.0.0.1 is especially useful if you're running some type of
local ZenML orchestrator, such as the k3d Kubeflow orchestrator or Docker
orchestrator, that cannot directly connect to the local ZenML server.

For example, to start the dashboard on port 9000 and have it listen
on all locally available interfaces on your machine, run:

```bash
zenml login --local --port 9000 --ip-address 0.0.0.0
```

Note that the above 0.0.0.0 IP address also exposes your ZenML dashboard
externally through your public interface. Alternatively, you can choose an
explicit IP address that is configured on one of your local interfaces, such as
the Docker bridge interface, which usually has the IP address `172.17.0.1`:

```bash
zenml login --local --port 9000 --ip-address 172.17.0.1
```

If you would like to take a look at the logs for the local ZenML server:

```bash
zenml logs
```

Connecting to a ZenML Server
----------------------------

The ZenML client can be [configured to connect to a local ZenML server, a remote
database or a remote ZenML server](https://docs.zenml.io/how-to/manage-zenml-server/connecting-to-zenml)
with the `zenml login` command.

To connect or re-connect to any ZenML server, if you know its URL, you can
simply run:

```bash
zenml login https://zenml.example.com:8080
```

Running `zenml login` without any arguments will check if your current ZenML
server session is still valid. If it is not, you will be prompted to log in
again. This is useful if you quickly want to refresh your CLI session when
the current session expires.

You can open the ZenML dashboard of your currently connected ZenML server using
the following command:

```bash
zenml server show
```

Note that if you have set your `AUTO_OPEN_DASHBOARD` environment variable to
`false` then this will not open the dashboard until you set it back to `true`.

The CLI can be authenticated to multiple ZenML servers at the same time,
even though it can only be connected to one server at a time. You can list
all the ZenML servers that the client is currently authenticated to by
running:

```bash
zenml server list
```

To disconnect from the current ZenML server and revert to using the local
default database, use the following command:

```bash
zenml logout
```

You can inspect the current ZenML configuration at any given time using the
following command:

```bash
zenml status
```

Example output:

```
$ zenml status
-----ZenML Client Status-----
Connected to a ZenML Pro server: `test-zenml-login` [16f8a35d-5c2f-44aa-a564-b34186fbf6d6]
  ZenML Pro Organization: My Organization
  ZenML Pro authentication: valid until 2024-10-26 10:18:51 CEST (in 20h37m9s)
  Dashboard: https://cloud.zenml.io/organizations/bf873af9-aaf9-4ad1-a08e-3dc6d920d590/tenants/16f8336d-5c2f-44aa-a534-b34186fbf6d6
  API: https://6784e58f-zenml.staging.cloudinfra.zenml.io
  Server status: 'available'
  Server authentication: never expires
  The active user is: 'user'
  The active stack is: 'default' (global)
Using configuration from: '/home/user/.config/zenml'
Local store files are located at: '/home/user/.config/zenml/local_stores'

-----Local ZenML Server Status-----
The local daemon server is running at: http://127.0.0.1:8237
```

When connecting to a ZenML server using the web login flow, you will be provided
with the option to `Trust this device`. If you opt out of it a 24-hour token
will be issued for the authentication service. If you opt-in, you will be
issued a 30-day token instead.

If you would like to see a list of all trusted devices, you can use:

```bash
zenml authorized-device list
```

or if you would like to get the details regarding a specific device,
you can use:

```bash
zenml authorized-device describe DEVICE_ID_OR_PREFIX
```

Alternatively, you can lock and unlock an authorized device by using the
following commands:

```bash
zenml authorized-device lock DEVICE_ID_OR_PREFIX
zenml authorized-device unlock DEVICE_ID_OR_PREFIX
```

Finally, you can remove an authorized device by using the `delete` command:

```bash
zenml authorized-device delete DEVICE_ID_OR_PREFIX
```

Secrets management
------------------

ZenML offers a way to [securely store secrets associated with your other
stack components and infrastructure](https://docs.zenml.io/deploying-zenml/deploying-zenml/secret-management).
A ZenML Secret is a collection or grouping of key-value pairs stored by the
ZenML secrets store. ZenML Secrets are identified by a unique name which
allows you to fetch or reference them in your pipelines and stacks.

Depending on how you set up and deployed ZenML, the secrets store keeps secrets
in the local database or uses the ZenML server your client is connected to:

* if you are using the default ZenML client settings, or if you connect your
ZenML client to a local ZenML server started with `zenml login --local`, the
secrets store is using the same local SQLite database as the rest of ZenML
* if you connect your ZenML client to a remote ZenML server, the
secrets are no longer managed on your local machine, but through the remote
server instead. Secrets are stored in whatever secrets store back-end the
remote server is configured to use. This can be a SQL database, one of the
managed cloud secrets management services, or even a custom back-end.

To create a secret, use the `create` command and pass the key-value pairs
as command-line arguments:

```bash
zenml secret create SECRET_NAME --key1=value1 --key2=value2 --key3=value3 ...

# Another option is to use the '--values' option and provide key-value pairs in either JSON or YAML format.
zenml secret create SECRET_NAME --values='{"key1":"value2","key2":"value2","key3":"value3"}'
```

Note that when using the previous command the keys and values will be preserved in your `bash_history` file, so
you may prefer to use the interactive `create` command instead:

```shell
zenml secret create SECRET_NAME -i
```

As an alternative to the interactive mode, also useful for values that
are long or contain newline or special characters, you can also use the special
`@` syntax to indicate to ZenML that the value needs to be read from a file:

```bash
zenml secret create SECRET_NAME \
   --aws_access_key_id=1234567890 \
   --aws_secret_access_key=abcdefghij \
   --aws_session_token=@/path/to/token.txt

# Alternatively for providing key-value pairs, you can utilize the '--values' option by specifying a file path containing
# key-value pairs in either JSON or YAML format.
zenml secret create SECRET_NAME --values=@/path/to/token.txt
```

To list all the secrets available, use the `list` command:

```bash
zenml secret list
```

To get the key-value pairs for a particular secret, use the `get` command:

```bash
zenml secret get SECRET_NAME
```

To update a secret, use the `update` command:

```bash
zenml secret update SECRET_NAME --key1=value1 --key2=value2 --key3=value3 ...

# Another option is to use the '--values' option and provide key-value pairs in either JSON or YAML format.
zenml secret update SECRET_NAME --values='{"key1":"value2","key2":"value2","key3":"value3"}'
```

Note that when using the previous command the keys and values will be preserved in your `bash_history` file, so
you may prefer to use the interactive `update` command instead:

```shell
zenml secret update SECRET_NAME -i
```

Finally, to delete a secret, use the `delete` command:

```bash
zenml secret delete SECRET_NAME
```

Secrets can be either private or public. Private secrets are only accessible
to the current user. Public secrets are accessible to all other users. By
default, secrets are public. To make a secret private, use the `--private` flag
in the `create` or `update` commands.

Auth management
---------------

Building and maintaining an MLOps workflow can involve numerous third-party
libraries and external services. In most cases, this ultimately presents a
challenge in configuring uninterrupted, secure access to infrastructure
resources. In ZenML, Service Connectors streamline this process by abstracting
away the complexity of authentication and help you connect your stack to your
resources. You can find the full docs on the ZenML service connectors
[here](https://docs.zenml.io/stacks/service-connectors/auth-management).

The ZenML CLI features a variety of commands to help you manage your service
connectors. First of all, to explore all the types of service connectors
available in ZenML, you can use the following commands:

```bash
# To get the complete list
zenml service-connector list-types

# To get the details regarding a single type
zenml service-connector describe-type
```

For each type of service connector, you will also see a list of supported
resource types. These types provide a way for organizing different resources
into logical classes based on the standard and/or protocol used to access them.
In addition to the resource types, each type will feature a different set of
authentication methods.

Once you decided which service connector to use, you can create it with the
`register` command as follows:

```bash
zenml service-connector register SERVICE_CONNECTOR_NAME \
    --type TYPE [--description DESCRIPTION] [--resource-type RESOURCE_TYPE] \
    [--auth-method AUTH_METHOD] ...
```

For more details on how to create a service connector, please refer to our
[docs](https://docs.zenml.io/stacks/service-connectors/auth-management).

To check if your service connector is registered properly, you can `verify` it.
By doing this, you can both check if it is configured correctly and also, you
can fetch the list of resources it has access to:

```bash
zenml service-connector verify SERVICE_CONNECTOR_NAME_ID_OR_PREFIX
```

Some service connectors come equipped with the capability of configuring
the clients and SDKs on your local machine with the credentials inferred from
your service connector. To use this functionality, simply use the `login`
command:

```bash
zenml service-connector login SERVICE_CONNECTOR_NAME_ID_OR_PREFIX
```

To list all the service connectors that you have registered, you can use:

```bash
zenml service-connector list
```

Moreover, if you would like to list all the resources accessible by your
service connectors, you can use the following command:

```bash
zenml service-connector list-resources [--resource-type RESOURCE_TYPE] /
    [--connector-type CONNECTOR_TYPE] ...
```

This command can possibly take a long time depending on the number of service
connectors you have registered. Consider using the right filters when you are
listing resources.

If you want to see the details about a specific service connector that you have
registered, you can use the `describe` command:

```bash
zenml service-connector describe SERVICE_CONNECTOR_NAME_ID_OR_PREFIX
```

You can update a registered service connector by using the `update` command.
Keep in mind that all service connector updates are validated before being
applied. If you want to disable this behavior please use the `--no-verify`
flag.

```bash
zenml service-connector update SERVICE_CONNECTOR_NAME_ID_OR_PREFIX ...
```

Finally, if you wish to remove a service connector, you can use the `delete`
command:

```bash
zenml service-connector delete SERVICE_CONNECTOR_NAME_ID_OR_PREFIX
```

Managing users
--------------

When using the ZenML service, you can manage permissions by managing users
using the CLI. If you want to create a new user or delete an existing one,
run either

```bash
zenml user create USER_NAME
zenml user delete USER_NAME
```

To see a list of all users, run:

```bash
zenml user list
```

For detail about the particular user, use the `describe` command. By default,
(without a specific user name passed in) it will describe the active user:

```bash
zenml user describe [USER_NAME]
```

If you want to update any properties of a specific user, you can use the
`update` command. Use the `--help` flag to get a full list of available
properties to update:

```bash
zenml user update --help
```

If you want to change the password of the current user account:

```bash
zenml user change-password --help
```

Service Accounts
----------------

ZenML supports the use of service accounts to authenticate clients to the
ZenML server using API keys. This is useful for automating tasks such as
running pipelines or deploying models.

To create a new service account, run:

```bash
zenml service-account create SERVICE_ACCOUNT_NAME
```

This command creates a service account and an API key for it. The API key is
displayed as part of the command output and cannot be retrieved later. You can
then use the issued API key to connect your ZenML client to the server with the
CLI:

```bash
zenml login https://... --api-key
```

or by setting the `ZENML_STORE_URL` and `ZENML_STORE_API_KEY` environment
variables when you set up your ZenML client for the first time:

```bash
export ZENML_STORE_URL=https://...
export ZENML_STORE_API_KEY=<API_KEY>
```

You don't need to run `zenml login` after setting these two environment
variables and can start interacting with your server right away.

To see all the service accounts you've created and their API keys, use the
following commands:

```bash
zenml service-account list
zenml service-account api-key <SERVICE_ACCOUNT_NAME> list
```

Additionally, the following command allows you to more precisely inspect one of
these service accounts and an API key:

```bash
zenml service-account describe <SERVICE_ACCOUNT_NAME>
zenml service-account api-key <SERVICE_ACCOUNT_NAME> describe <API_KEY_NAME>
```

API keys don't have an expiration date. For increased security, we recommend
that you regularly rotate the API keys to prevent unauthorized access to your
ZenML server. You can do this with the ZenML CLI:

```bash
zenml service-account api-key <SERVICE_ACCOUNT_NAME> rotate <API_KEY_NAME>
```

Running this command will create a new API key and invalidate the old one. The
new API key is displayed as part of the command output and cannot be retrieved
later. You can then use the new API key to connect your ZenML client to the
server just as described above.

When rotating an API key, you can also configure a retention period for the old
API key. This is useful if you need to keep the old API key for a while to
ensure that all your workloads have been updated to use the new API key. You can
do this with the `--retain` flag. For example, to rotate an API key and keep the
old one for 60 minutes, you can run the following command:

```bash
zenml service-account api-key <SERVICE_ACCOUNT_NAME> rotate <API_KEY_NAME> \
      --retain 60
```

For increased security, you can deactivate a service account or an API key using
one of the following commands:

```
zenml service-account update <SERVICE_ACCOUNT_NAME> --active false
zenml service-account api-key <SERVICE_ACCOUNT_NAME> update <API_KEY_NAME> \
      --active false
```

Deactivating a service account or an API key will prevent it from being used to
authenticate and has immediate effect on all workloads that use it.

To permanently delete an API key for a service account, use the following
command:

```bash
zenml service-account api-key <SERVICE_ACCOUNT_NAME> delete <API_KEY_NAME>
```

Managing Code Repositories
--------------------------

[Code repositories](https://docs.zenml.io/user-guides/production-guide/connect-code-repository)
enable ZenML to keep track of the code version that you use for your pipeline
runs. Additionally, running a pipeline which is tracked in a registered code
repository can decrease the time it takes Docker to build images for
containerized stack components.

To register a code repository, use the following CLI
command:

```shell
zenml code-repository register <NAME> --type=<CODE_REPOSITORY_TYPE] \
   [--CODE_REPOSITORY_OPTIONS]
```

ZenML currently supports code repositories of type `github` and `gitlab`, but
you can also use your custom code repository implementation by passing the
type `custom` and a source of your repository class.

```shell
zenml code-repository register <NAME> --type=custom \
   --source=<CODE_REPOSITORY_SOURCE> [--CODE_REPOSITORY_OPTIONS]
```

The `CODE_REPOSITORY_OPTIONS` depend on the configuration necessary for the
type of code repository that you're using.

If you want to list your registered code repositories, run:

```shell
zenml code-repository list
```

You can delete one of your registered code repositories like this:

```shell
zenml code-repository delete <REPOSITORY_NAME_OR_ID>
```

Building an image without Runs
------------------------------

To build or run a pipeline from the CLI, you need to know the source path of
your pipeline. Let's imagine you have defined your pipeline in a python file
called `run.py` like this:

```python
from zenml import pipeline

@pipeline
def my_pipeline(...):
   # Connect your pipeline steps here
   pass
```

The source path of your pipeline will be `run.my_pipeline`. In a generalized
way, this will be `<MODULE_PATH>.<PIPELINE_FUNCTION_NAME>`. If the python file
defining the pipeline is not in your current directory, the module path consists
of the full path to the file, separated by dots, e.g.
`some_directory.some_file.my_pipeline`.

To [build Docker images for your pipeline](https://docs.zenml.io/concepts/containerization)
without actually running the pipeline, use:

```bash
zenml pipeline build <PIPELINE_SOURCE_PATH>
```

To specify settings for the Docker builds, use the `--config/-c` option of the
command. For more information about the structure of this configuration file,
check out the `zenml.pipelines.base_pipeline.BasePipeline.build(...)` method.

```bash
zenml pipeline build <PIPELINE_SOURCE_PATH> --config=<PATH_TO_CONFIG_YAML>
```

If you want to build the pipeline for a stack other than your current active
stack, use the `--stack` option.

```bash
zenml pipeline build <PIPELINE_SOURCE_PATH> --stack=<STACK_ID_OR_NAME>
```

To run a pipeline that was previously registered, use:

```bash
zenml pipeline run <PIPELINE_SOURCE_PATH>
```

To specify settings for the pipeline, use the `--config/-c` option of the
command. For more information about the structure of this configuration file,
check out the `zenml.pipelines.base_pipeline.BasePipeline.run(...)` method.

```bash
zenml pipeline run <PIPELINE_SOURCE_PATH> --config=<PATH_TO_CONFIG_YAML>
```

If you want to run the pipeline on a stack different than your current active
stack, use the `--stack` option.

```bash
zenml pipeline run <PIPELINE_SOURCE_PATH> --stack=<STACK_ID_OR_NAME>
```

If you want to create a run template based on your pipeline that can later be used to trigger a run either from the dashboard or through an HTTP request:

```bash
zenml pipeline create-run-template <PIPELINE_SOURCE_PATH> \
   --name=<TEMPLATE_NAME>
```

To specify a config file, use the `--config/-c` option. If you would like to use a different stack than the active one, use the `--stack` option.

```bash
zenml pipeline create-run-template <PIPELINE_SOURCE_PATH> \
   --name=<TEMPLATE_NAME> \
   --config=<PATH_TO_CONFIG_YAML> \
   --stack=<STACK_ID_OR_NAME>
```

Tagging your resources with ZenML
---------------------------------

When you are using ZenML, you can [use tags to organize and categorize your
assets](https://docs.zenml.io/concepts/tags).
This way, you can streamline your workflows and enhance the discoverability of
your resources more easily.

Currently, you can use tags with artifacts, models and their versions:

```bash
# Tag the artifact
zenml artifact update ARTIFACT_NAME -t TAG_NAME

# Tag the artifact version
zenml artifact version update ARTIFACT_NAME ARTIFACT_VERSION -t TAG_NAME

# Tag an existing model
zenml model update MODEL_NAME --tag TAG_NAME

# Tag a specific model version
zenml model version update MODEL_NAME VERSION_NAME --tag TAG_NAME
```

Besides these interactions, you can also create a new tag by using the
`register` command:

```bash
zenml tag register -n TAG_NAME [-c COLOR]
```

If you would like to list all the tags that you have, you can use the command:

```bash
zenml tag list
```

To update the properties of a specific tag, you can use the `update` subcommand:

```bash
zenml tag update TAG_NAME_OR_ID [-n NEW_NAME] [-c NEW_COLOR]
```

Finally, in order to delete a tag, you can execute:

```bash
zenml tag delete TAG_NAME_OR_ID
```

Managing the Global Configuration
---------------------------------

The ZenML global configuration CLI commands cover options such as enabling or
disabling the collection of anonymous usage statistics, changing the logging
verbosity.

In order to help us better understand how the community uses ZenML, the library
reports anonymized usage statistics. You can always opt out by using the CLI
command:

```bash
zenml analytics opt-out
```

If you want to opt back in, use the following command:

```bash
zenml analytics opt-in
```

The verbosity of the ZenML client output can be configured using the
``zenml logging`` command. For example, to set the verbosity to DEBUG, run:

```bash
zenml logging set-verbosity DEBUG
```
"""

from zenml.cli.version import *  # noqa
from zenml.cli.annotator import *  # noqa
from zenml.cli.artifact import *  # noqa
from zenml.cli.authorized_device import *  # noqa
from zenml.cli.base import *  # noqa
from zenml.cli.code_repository import *  # noqa
from zenml.cli.config import *  # noqa
from zenml.cli.downgrade import *  # noqa
from zenml.cli.feature import *  # noqa
from zenml.cli.integration import *  # noqa
from zenml.cli.login import *
from zenml.cli.model import *  # noqa
from zenml.cli.model_registry import *  # noqa
from zenml.cli.pipeline import *  # noqa
from zenml.cli.secret import *  # noqa
from zenml.cli.served_model import *  # noqa
from zenml.cli.server import *  # noqa
from zenml.cli.service_accounts import *  # noqa
from zenml.cli.service_connectors import *  # noqa
from zenml.cli.stack import *  # noqa
from zenml.cli.stack_components import *  # noqa
from zenml.cli.user_management import *  # noqa
from zenml.cli.project import *  # noqa
from zenml.cli.tag import *  # noqa



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for run templates."""

from typing import Optional, Union
from uuid import UUID

from fastapi import (
    APIRouter,
    Depends,
    Security,
)

from zenml.analytics.enums import AnalyticsEvent
from zenml.analytics.utils import track_handler
from zenml.config.pipeline_run_configuration import PipelineRunConfiguration
from zenml.constants import (
    API,
    RUN_TEMPLATE_TRIGGERS_FEATURE_NAME,
    RUN_TEMPLATES,
    VERSION_1,
)
from zenml.models import (
    Page,
    PipelineRunResponse,
    RunTemplateFilter,
    RunTemplateRequest,
    RunTemplateResponse,
    RunTemplateUpdate,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.feature_gate.endpoint_utils import (
    check_entitlement,
)
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import verify_permission
from zenml.zen_server.routers.projects_endpoints import workspace_router
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    server_config,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + RUN_TEMPLATES,
    tags=["run_templates"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.post(
    "/{project_name_or_id}" + RUN_TEMPLATES,
    responses={401: error_response, 409: error_response, 422: error_response},
    deprecated=True,
    tags=["run_templates"],
)
@async_fastapi_endpoint_wrapper
def create_run_template(
    run_template: RunTemplateRequest,
    project_name_or_id: Optional[Union[str, UUID]] = None,
    _: AuthContext = Security(authorize),
) -> RunTemplateResponse:
    """Create a run template.

    Args:
        run_template: Run template to create.
        project_name_or_id: Optional name or ID of the project.

    Returns:
        The created run template.
    """
    if project_name_or_id:
        project = zen_store().get_project(project_name_or_id)
        run_template.project = project.id

    return verify_permissions_and_create_entity(
        request_model=run_template,
        create_method=zen_store().create_run_template,
    )


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.get(
    "/{project_name_or_id}" + RUN_TEMPLATES,
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
    tags=["run_templates"],
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def list_run_templates(
    filter_model: RunTemplateFilter = Depends(
        make_dependable(RunTemplateFilter)
    ),
    project_name_or_id: Optional[Union[str, UUID]] = None,
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[RunTemplateResponse]:
    """Get a page of run templates.

    Args:
        filter_model: Filter model used for pagination, sorting,
            filtering.
        project_name_or_id: Optional name or ID of the project.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        Page of run templates.
    """
    if project_name_or_id:
        filter_model.project = project_name_or_id

    return verify_permissions_and_list_entities(
        filter_model=filter_model,
        resource_type=ResourceType.RUN_TEMPLATE,
        list_method=zen_store().list_run_templates,
        hydrate=hydrate,
    )


@router.get(
    "/{template_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def get_run_template(
    template_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> RunTemplateResponse:
    """Get a run template.

    Args:
        template_id: ID of the run template to get.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        The run template.
    """
    return verify_permissions_and_get_entity(
        id=template_id,
        get_method=zen_store().get_run_template,
        hydrate=hydrate,
    )


@router.put(
    "/{template_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def update_run_template(
    template_id: UUID,
    update: RunTemplateUpdate,
    _: AuthContext = Security(authorize),
) -> RunTemplateResponse:
    """Update a run template.

    Args:
        template_id: ID of the run template to get.
        update: The updates to apply.

    Returns:
        The updated run template.
    """
    return verify_permissions_and_update_entity(
        id=template_id,
        update_model=update,
        get_method=zen_store().get_run_template,
        update_method=zen_store().update_run_template,
    )


@router.delete(
    "/{template_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_run_template(
    template_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Delete a run template.

    Args:
        template_id: ID of the run template to delete.
    """
    verify_permissions_and_delete_entity(
        id=template_id,
        get_method=zen_store().get_run_template,
        delete_method=zen_store().delete_run_template,
    )


if server_config().workload_manager_enabled:

    @router.post(
        "/{template_id}/runs",
        responses={
            401: error_response,
            404: error_response,
            422: error_response,
            429: error_response,
        },
    )
    @async_fastapi_endpoint_wrapper
    def create_template_run(
        template_id: UUID,
        config: Optional[PipelineRunConfiguration] = None,
        auth_context: AuthContext = Security(authorize),
    ) -> PipelineRunResponse:
        """Run a pipeline from a template.

        Args:
            template_id: The ID of the template.
            config: Configuration for the pipeline run.
            auth_context: Authentication context.

        Returns:
            The created pipeline run.
        """
        from zenml.zen_server.template_execution.utils import (
            run_template,
        )

        with track_handler(
            event=AnalyticsEvent.EXECUTED_RUN_TEMPLATE,
        ) as analytics_handler:
            template = verify_permissions_and_get_entity(
                id=template_id,
                get_method=zen_store().get_run_template,
                hydrate=True,
            )
            analytics_handler.metadata = {
                "project_id": template.project_id,
            }

            verify_permission(
                resource_type=ResourceType.PIPELINE_DEPLOYMENT,
                action=Action.CREATE,
                project_id=template.project_id,
            )
            verify_permission(
                resource_type=ResourceType.PIPELINE_RUN,
                action=Action.CREATE,
                project_id=template.project_id,
            )
            check_entitlement(feature=RUN_TEMPLATE_TRIGGERS_FEATURE_NAME)

            return run_template(
                template=template,
                auth_context=auth_context,
                run_config=config,
            )



"""Add hidden option for templates [6611d4bcc95b].

Revision ID: 6611d4bcc95b
Revises: 0.80.1
Create Date: 2025-04-07 10:29:26.130057

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6611d4bcc95b"
down_revision = "0.80.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.add_column(sa.Column("hidden", sa.Boolean(), nullable=True))

    op.execute(
        """
        UPDATE `run_template` 
        SET hidden = 0
        """
    )

    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.alter_column(
            "hidden",
            existing_type=sa.Boolean(),
            nullable=False,
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.drop_column("hidden")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Neptune experiment tracker flavor."""

__all__ = [
    "NeptuneExperimentTrackerConfig",
    "NeptuneExperimentTrackerFlavor",
    "NeptuneExperimentTrackerSettings",
]

from typing import TYPE_CHECKING, Optional, Set, Type

from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.experiment_trackers.base_experiment_tracker import (
    BaseExperimentTrackerConfig,
    BaseExperimentTrackerFlavor,
)
from zenml.integrations.neptune import NEPTUNE_MODEL_EXPERIMENT_TRACKER_FLAVOR
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.neptune.experiment_trackers import (
        NeptuneExperimentTracker,
    )


class NeptuneExperimentTrackerConfig(BaseExperimentTrackerConfig):
    """Config for the Neptune experiment tracker.

    If attributes are left as None, the neptune.init_run() method
    will try to find the relevant values in the environment
    """

    project: Optional[str] = Field(
        None,
        description="Name of the Neptune project you want to log the metadata to.",
    )
    api_token: Optional[str] = SecretField(
        default=None, description="Your Neptune API token for authentication."
    )


class NeptuneExperimentTrackerSettings(BaseSettings):
    """Settings for the Neptune experiment tracker."""

    tags: Set[str] = Field(
        default_factory=set, description="Tags for the Neptune run."
    )


class NeptuneExperimentTrackerFlavor(BaseExperimentTrackerFlavor):
    """Class for the `NeptuneExperimentTrackerFlavor`."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return NEPTUNE_MODEL_EXPERIMENT_TRACKER_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/experiment_tracker/neptune.png"

    @property
    def config_class(self) -> Type[NeptuneExperimentTrackerConfig]:
        """Returns `NeptuneExperimentTrackerConfig` config class.

        Returns:
                The config class.
        """
        return NeptuneExperimentTrackerConfig

    @property
    def implementation_class(self) -> Type["NeptuneExperimentTracker"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.neptune.experiment_trackers import (
            NeptuneExperimentTracker,
        )

        return NeptuneExperimentTracker



"""Added service table [0701da9951a0].

Revision ID: 0701da9951a0
Revises: 0.56.1
Create Date: 2024-03-25 12:24:32.928543

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = "0701da9951a0"
down_revision = "0.56.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # If the tables already exist, skip this migration.
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn.engine)
    tables = inspector.get_table_names()
    if "service" in tables:
        return

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "service",
        sa.Column(
            "workspace_id", sqlmodel.sql.sqltypes.GUID(), nullable=False
        ),
        sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
        sa.Column("service_source", sa.TEXT(), nullable=True),
        sa.Column("service_type", sa.TEXT(), nullable=False),
        sa.Column("type", sa.TEXT(), nullable=False),
        sa.Column("flavor", sa.TEXT(), nullable=False),
        sa.Column("admin_state", sa.TEXT(), nullable=True),
        sa.Column("state", sa.TEXT(), nullable=True),
        sa.Column("prediction_url", sa.TEXT(), nullable=True),
        sa.Column("health_check_url", sa.TEXT(), nullable=True),
        sa.Column("pipeline_name", sa.TEXT(), nullable=True),
        sa.Column("pipeline_step_name", sa.TEXT(), nullable=True),
        sa.Column(
            "model_version_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
        ),
        sa.Column(
            "pipeline_run_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
        ),
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("labels", sa.LargeBinary(), nullable=True),
        sa.Column("config", sa.LargeBinary(), nullable=False),
        sa.Column("status", sa.LargeBinary(), nullable=True),
        sa.Column("endpoint", sa.LargeBinary(), nullable=True),
        sa.ForeignKeyConstraint(
            ["model_version_id"],
            ["model_version.id"],
            name="fk_service_model_version_id_model_version",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["pipeline_run_id"],
            ["pipeline_run.id"],
            name="fk_service_pipeline_run_id_pipeline_run",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_service_user_id_user",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["workspace_id"],
            ["workspace.id"],
            name="fk_service_workspace_id_workspace",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("service")
    # ### end Alembic commands ###



"""Update build filtering [d26471b6fe8f].

Revision ID: d26471b6fe8f
Revises: 4c41c0ca42db
Create Date: 2023-03-07 17:31:39.243348

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "d26471b6fe8f"
down_revision = "4c41c0ca42db"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "zenml_version",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column("contains_code", sa.Boolean(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "python_version",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column(
                "checksum", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
    op.execute("UPDATE pipeline_build SET contains_code = 1")
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.alter_column(
            "contains_code",
            nullable=False,
            existing_type=sa.Boolean(),
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.drop_column("zenml_version")
        batch_op.drop_column("checksum")
        batch_op.drop_column("python_version")
        batch_op.drop_column("contains_code")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""BentoML model deployer flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.integrations.bentoml import BENTOML_MODEL_DEPLOYER_FLAVOR
from zenml.model_deployers.base_model_deployer import (
    BaseModelDeployerConfig,
    BaseModelDeployerFlavor,
)

if TYPE_CHECKING:
    from zenml.integrations.bentoml.model_deployers import BentoMLModelDeployer


class BentoMLModelDeployerConfig(BaseModelDeployerConfig):
    """Configuration for the BentoMLModelDeployer."""

    service_path: str = Field(
        "",
        description="Path to the BentoML service directory. "
        "If not provided, a default service path will be used.",
    )


class BentoMLModelDeployerFlavor(BaseModelDeployerFlavor):
    """Flavor for the BentoML model deployer."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            Name of the flavor.
        """
        return BENTOML_MODEL_DEPLOYER_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/model_deployer/bentoml.png"

    @property
    def config_class(self) -> Type[BentoMLModelDeployerConfig]:
        """Returns `BentoMLModelDeployerConfig` config class.

        Returns:
                The config class.
        """
        return BentoMLModelDeployerConfig

    @property
    def implementation_class(self) -> Type["BentoMLModelDeployer"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.bentoml.model_deployers import (
            BentoMLModelDeployer,
        )

        return BentoMLModelDeployer



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes step operator flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.constants import KUBERNETES_CLUSTER_RESOURCE_TYPE
from zenml.integrations.kubernetes import KUBERNETES_STEP_OPERATOR_FLAVOR
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.models import ServiceConnectorRequirements
from zenml.step_operators import BaseStepOperatorConfig, BaseStepOperatorFlavor

if TYPE_CHECKING:
    from zenml.integrations.kubernetes.step_operators import (
        KubernetesStepOperator,
    )


class KubernetesStepOperatorSettings(BaseSettings):
    """Settings for the Kubernetes step operator.

    Configuration options for individual step execution on Kubernetes.
    Field descriptions are defined inline using Field() descriptors.
    """

    pod_settings: Optional[KubernetesPodSettings] = Field(
        default=None,
        description="Pod configuration for step execution containers.",
    )
    service_account_name: Optional[str] = Field(
        default=None,
        description="Kubernetes service account for step pods. Uses default account if not specified.",
    )
    privileged: bool = Field(
        default=False,
        description="Whether to run step containers in privileged mode with extended permissions.",
    )
    pod_startup_timeout: int = Field(
        default=600,
        description="Maximum seconds to wait for step pods to start. Default is 10 minutes.",
    )
    pod_failure_max_retries: int = Field(
        default=3,
        description="Maximum retry attempts when step pods fail to start.",
    )
    pod_failure_retry_delay: int = Field(
        default=10,
        description="Delay in seconds between pod failure retry attempts.",
    )
    pod_failure_backoff: float = Field(
        default=1.0,
        description="Exponential backoff factor for retry delays. Values > 1.0 increase delay with each retry.",
    )


class KubernetesStepOperatorConfig(
    BaseStepOperatorConfig, KubernetesStepOperatorSettings
):
    """Configuration for the Kubernetes step operator.

    Defines cluster connection and execution settings.
    Field descriptions are defined inline using Field() descriptors.
    """

    kubernetes_namespace: str = Field(
        default="zenml",
        description="Kubernetes namespace for step execution. Must be a valid namespace name.",
    )
    incluster: bool = Field(
        default=False,
        description="Whether to execute within the same cluster as the orchestrator. "
        "Requires appropriate pod creation permissions.",
    )
    kubernetes_context: Optional[str] = Field(
        default=None,
        description="Kubernetes context name for cluster connection. "
        "Ignored when using service connectors or in-cluster execution.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return False


class KubernetesStepOperatorFlavor(BaseStepOperatorFlavor):
    """Kubernetes step operator flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return KUBERNETES_STEP_OPERATOR_FLAVOR

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type=KUBERNETES_CLUSTER_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/step_operator/kubernetes.png"

    @property
    def config_class(self) -> Type[KubernetesStepOperatorConfig]:
        """Returns `KubernetesStepOperatorConfig` config class.

        Returns:
                The config class.
        """
        return KubernetesStepOperatorConfig

    @property
    def implementation_class(self) -> Type["KubernetesStepOperator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.kubernetes.step_operators import (
            KubernetesStepOperator,
        )

        return KubernetesStepOperator



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Amazon SageMaker orchestrator flavor."""

from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from pydantic import Field, model_validator

from zenml.config.base_settings import BaseSettings
from zenml.integrations.aws import (
    AWS_RESOURCE_TYPE,
    AWS_SAGEMAKER_STEP_OPERATOR_FLAVOR,
)
from zenml.models import ServiceConnectorRequirements
from zenml.orchestrators import BaseOrchestratorConfig
from zenml.orchestrators.base_orchestrator import BaseOrchestratorFlavor
from zenml.utils import deprecation_utils
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.aws.orchestrators import SagemakerOrchestrator

DEFAULT_TRAINING_INSTANCE_TYPE = "ml.m5.xlarge"
DEFAULT_PROCESSING_INSTANCE_TYPE = "ml.t3.medium"
DEFAULT_OUTPUT_DATA_S3_MODE = "EndOfJob"


class SagemakerOrchestratorSettings(BaseSettings):
    """Settings for the Sagemaker orchestrator."""

    synchronous: bool = Field(
        True,
        description="Controls whether pipeline execution blocks the client. If True, "
        "the client waits until all steps complete before returning. If False, "
        "returns immediately and executes asynchronously. Useful for long-running "
        "production pipelines where you don't want to maintain a connection",
    )

    instance_type: Optional[str] = Field(
        None,
        description="AWS EC2 instance type for step execution. Must be a valid "
        "SageMaker-supported instance type. Examples: 'ml.t3.medium' (2 vCPU, 4GB RAM), "
        "'ml.m5.xlarge' (4 vCPU, 16GB RAM), 'ml.p3.2xlarge' (8 vCPU, 61GB RAM, 1 GPU). "
        "Defaults to ml.m5.xlarge for training steps or ml.t3.medium for processing steps",
    )
    execution_role: Optional[str] = Field(
        None,
        description="IAM role ARN for SageMaker step execution permissions. Must have "
        "necessary policies attached (SageMakerFullAccess, S3 access, etc.). "
        "Example: 'arn:aws:iam::123456789012:role/SageMakerExecutionRole'. "
        "If not provided, uses the default SageMaker execution role",
    )
    volume_size_in_gb: int = Field(
        30,
        description="EBS volume size in GB for step execution storage. Must be between "
        "1-16384 GB. Used for temporary files, model artifacts, and data processing. "
        "Larger volumes needed for big datasets or model training. Example: 30 for "
        "small jobs, 100+ for large ML training jobs",
    )
    max_runtime_in_seconds: int = Field(
        86400,  # 24 hours
        description="Maximum execution time in seconds before job termination. Must be "
        "between 1-432000 seconds (5 days). Used to prevent runaway jobs and control costs. "
        "Examples: 3600 (1 hour), 86400 (24 hours), 259200 (3 days). "
        "Consider your longest expected step duration",
    )
    tags: Dict[str, str] = Field(
        default_factory=dict,
        description="Tags to apply to the Processor/Estimator assigned to the step. "
        "Example: {'Environment': 'Production', 'Project': 'MLOps'}",
    )
    pipeline_tags: Dict[str, str] = Field(
        default_factory=dict,
        description="Tags to apply to the pipeline via the "
        "sagemaker.workflow.pipeline.Pipeline.create method. Example: "
        "{'Environment': 'Production', 'Project': 'MLOps'}",
    )
    keep_alive_period_in_seconds: Optional[int] = Field(
        300,  # 5 minutes
        description="The time in seconds after which the provisioned instance "
        "will be terminated if not used. This is only applicable for "
        "TrainingStep type.",
    )
    use_training_step: Optional[bool] = Field(
        None,
        description="Whether to use the TrainingStep type. It is not possible "
        "to use TrainingStep type if the `output_data_s3_uri` is set to "
        "Dict[str, str] or if the `output_data_s3_mode` != 'EndOfJob'.",
    )

    processor_args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Arguments that are directly passed to the SageMaker "
        "Processor for a specific step, allowing for overriding the default "
        "settings provided when configuring the component. Example: "
        "{'instance_count': 2, 'base_job_name': 'my-processing-job'}",
    )
    estimator_args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Arguments that are directly passed to the SageMaker "
        "Estimator for a specific step, allowing for overriding the default "
        "settings provided when configuring the component. Example: "
        "{'train_instance_count': 2, 'train_max_run': 3600}",
    )
    environment: Dict[str, str] = Field(
        default_factory=dict,
        description="Environment variables to pass to the container. "
        "Example: {'LOG_LEVEL': 'INFO', 'DEBUG_MODE': 'False'}",
    )

    input_data_s3_mode: str = Field(
        "File",
        description="How data is made available to the container. "
        "Two possible input modes: File, Pipe.",
    )
    input_data_s3_uri: Optional[Union[str, Dict[str, str]]] = Field(
        default=None,
        union_mode="left_to_right",
        description="S3 URI where data is located if not locally. Example string: "
        "'s3://my-bucket/my-data/train'. Example dict: "
        "{'training': 's3://bucket/train', 'validation': 's3://bucket/val'}",
    )

    output_data_s3_mode: str = Field(
        DEFAULT_OUTPUT_DATA_S3_MODE,
        description="How data is uploaded to the S3 bucket. "
        "Two possible output modes: EndOfJob, Continuous.",
    )
    output_data_s3_uri: Optional[Union[str, Dict[str, str]]] = Field(
        default=None,
        union_mode="left_to_right",
        description="S3 URI where data is uploaded after or during processing run. "
        "Example string: 's3://my-bucket/my-data/output'. Example dict: "
        "{'output_one': 's3://bucket/out1', 'output_two': 's3://bucket/out2'}",
    )
    processor_role: Optional[str] = Field(
        None,
        description="DEPRECATED: use `execution_role` instead. "
        "The IAM role to use for the step execution.",
    )
    processor_tags: Optional[Dict[str, str]] = Field(
        None,
        description="DEPRECATED: use `tags` instead. "
        "Tags to apply to the Processor assigned to the step.",
    )
    _deprecation_validator = deprecation_utils.deprecate_pydantic_attributes(
        ("processor_role", "execution_role"), ("processor_tags", "tags")
    )

    @model_validator(mode="before")
    def validate_model(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        """Check if model is configured correctly.

        Args:
            data: The model data.

        Returns:
            The validated model data.

        Raises:
            ValueError: If the model is configured incorrectly.
        """
        use_training_step = data.get("use_training_step", True)
        output_data_s3_uri = data.get("output_data_s3_uri", None)
        output_data_s3_mode = data.get(
            "output_data_s3_mode", DEFAULT_OUTPUT_DATA_S3_MODE
        )
        if use_training_step and (
            isinstance(output_data_s3_uri, dict)
            or (
                isinstance(output_data_s3_uri, str)
                and (output_data_s3_mode != DEFAULT_OUTPUT_DATA_S3_MODE)
            )
        ):
            raise ValueError(
                "`use_training_step=True` is not supported when `output_data_s3_uri` is a dict or "
                f"when `output_data_s3_mode` is not '{DEFAULT_OUTPUT_DATA_S3_MODE}'."
            )
        instance_type = data.get("instance_type", None)
        if instance_type is None:
            if use_training_step:
                data["instance_type"] = DEFAULT_TRAINING_INSTANCE_TYPE
            else:
                data["instance_type"] = DEFAULT_PROCESSING_INSTANCE_TYPE
        return data


class SagemakerOrchestratorConfig(
    BaseOrchestratorConfig, SagemakerOrchestratorSettings
):
    """Config for the Sagemaker orchestrator.

    There are three ways to authenticate to AWS:
    - By connecting a `ServiceConnector` to the orchestrator,
    - By configuring explicit AWS credentials `aws_access_key_id`,
        `aws_secret_access_key`, and optional `aws_auth_role_arn`,
    - If none of the above are provided, unspecified credentials will be
        loaded from the default AWS config.
    """

    execution_role: str = Field(
        ..., description="The IAM role ARN to use for the pipeline."
    )
    scheduler_role: Optional[str] = Field(
        None,
        description="The ARN of the IAM role that will be assumed by "
        "the EventBridge service to launch Sagemaker pipelines. "
        "Required for scheduled pipelines.",
    )
    aws_access_key_id: Optional[str] = SecretField(
        default=None,
        description="The AWS access key ID to use to authenticate to AWS. "
        "If not provided, the value from the default AWS config will be used.",
    )
    aws_secret_access_key: Optional[str] = SecretField(
        default=None,
        description="The AWS secret access key to use to authenticate to AWS. "
        "If not provided, the value from the default AWS config will be used.",
    )
    aws_profile: Optional[str] = Field(
        None,
        description="The AWS profile to use for authentication if not using "
        "service connectors or explicit credentials. If not provided, the "
        "default profile will be used.",
    )
    aws_auth_role_arn: Optional[str] = Field(
        None,
        description="The ARN of an intermediate IAM role to assume when "
        "authenticating to AWS.",
    )
    region: Optional[str] = Field(
        None,
        description="The AWS region where the processing job will be run. "
        "If not provided, the value from the default AWS config will be used.",
    )
    bucket: Optional[str] = Field(
        None,
        description="Name of the S3 bucket to use for storing artifacts "
        "from the job run. If not provided, a default bucket will be created "
        "based on the following format: 'sagemaker-{region}-{aws-account-id}'.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return self.synchronous

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return True


class SagemakerOrchestratorFlavor(BaseOrchestratorFlavor):
    """Flavor for the Sagemaker orchestrator."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return AWS_SAGEMAKER_STEP_OPERATOR_FLAVOR

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(resource_type=AWS_RESOURCE_TYPE)

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/sagemaker.png"

    @property
    def config_class(self) -> Type[SagemakerOrchestratorConfig]:
        """Returns SagemakerOrchestratorConfig config class.

        Returns:
            The config class.
        """
        return SagemakerOrchestratorConfig

    @property
    def implementation_class(self) -> Type["SagemakerOrchestrator"]:
        """Implementation class.

        Returns:
            The implementation class.
        """
        from zenml.integrations.aws.orchestrators import SagemakerOrchestrator

        return SagemakerOrchestrator



"""Release [0.65.0].

Revision ID: 0.65.0
Revises: bf2120261b5a
Create Date: 2024-08-28 15:42:13.858104

"""

# revision identifiers, used by Alembic.
revision = "0.65.0"
down_revision = "bf2120261b5a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Zen Server API.

To run this file locally, execute:

    ```
    uvicorn zenml.zen_server.zen_server_api:app --reload
    ```
"""

import logging
import os
from asyncio.log import logger
from genericpath import isfile
from typing import Any, List

from anyio import to_thread
from fastapi import FastAPI, HTTPException, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import ORJSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.responses import (
    FileResponse,
    RedirectResponse,
)

import zenml
from zenml.constants import (
    API,
    HEALTH,
    READY,
)
from zenml.enums import AuthScheme
from zenml.models import ServerDeploymentType
from zenml.service_connectors.service_connector_registry import (
    service_connector_registry,
)
from zenml.zen_server.cloud_utils import send_pro_workspace_status_update
from zenml.zen_server.exceptions import error_detail
from zenml.zen_server.middleware import add_middlewares
from zenml.zen_server.routers import (
    actions_endpoints,
    artifact_endpoint,
    artifact_version_endpoints,
    auth_endpoints,
    code_repositories_endpoints,
    devices_endpoints,
    event_source_endpoints,
    flavors_endpoints,
    logs_endpoints,
    model_versions_endpoints,
    models_endpoints,
    pipeline_builds_endpoints,
    pipeline_deployments_endpoints,
    pipelines_endpoints,
    plugin_endpoints,
    projects_endpoints,
    run_metadata_endpoints,
    run_templates_endpoints,
    runs_endpoints,
    schedule_endpoints,
    secrets_endpoints,
    server_endpoints,
    service_accounts_endpoints,
    service_connectors_endpoints,
    service_endpoints,
    stack_components_endpoints,
    stack_deployment_endpoints,
    stacks_endpoints,
    steps_endpoints,
    tag_resource_endpoints,
    tags_endpoints,
    triggers_endpoints,
    users_endpoints,
    webhook_endpoints,
)
from zenml.zen_server.secure_headers import (
    initialize_secure_headers,
)
from zenml.zen_server.utils import (
    cleanup_request_manager,
    initialize_feature_gate,
    initialize_memcache,
    initialize_plugins,
    initialize_rbac,
    initialize_request_manager,
    initialize_run_template_executor,
    initialize_workload_manager,
    initialize_zen_store,
    run_template_executor,
    server_config,
    start_event_loop_lag_monitor,
    stop_event_loop_lag_monitor,
)

DASHBOARD_DIRECTORY = "dashboard"


def relative_path(rel: str) -> str:
    """Get the absolute path of a path relative to the ZenML server module.

    Args:
        rel: Relative path.

    Returns:
        Absolute path.
    """
    return os.path.join(os.path.dirname(__file__), rel)


app = FastAPI(
    title="ZenML",
    version=zenml.__version__,
    root_path=server_config().root_url_path,
    default_response_class=ORJSONResponse,
)

add_middlewares(app)


# Customize the default request validation handler that comes with FastAPI
# to return a JSON response that matches the ZenML API spec.
@app.exception_handler(RequestValidationError)
def validation_exception_handler(
    request: Any, exc: Exception
) -> ORJSONResponse:
    """Custom validation exception handler.

    Args:
        request: The request.
        exc: The exception.

    Returns:
        The error response formatted using the ZenML API conventions.
    """
    return ORJSONResponse(error_detail(exc, ValueError), status_code=422)


@app.on_event("startup")
async def initialize() -> None:
    """Initialize the ZenML server."""
    cfg = server_config()
    # Set the maximum number of worker threads
    to_thread.current_default_thread_limiter().total_tokens = (
        cfg.thread_pool_size
    )
    # IMPORTANT: these need to be run before the fastapi app starts, to avoid
    # race conditions
    await initialize_request_manager()
    initialize_zen_store()
    service_connector_registry.register_builtin_service_connectors()
    initialize_rbac()
    initialize_feature_gate()
    initialize_workload_manager()
    initialize_run_template_executor()
    initialize_plugins()
    initialize_secure_headers()
    initialize_memcache(cfg.memcache_max_capacity, cfg.memcache_default_expiry)
    if cfg.deployment_type == ServerDeploymentType.CLOUD:
        # Send a workspace status update to the Cloud API to indicate that the
        # ZenML server is running or to update the version and server URL.
        send_pro_workspace_status_update()

    if logger.isEnabledFor(logging.DEBUG):
        start_event_loop_lag_monitor()


@app.on_event("shutdown")
async def shutdown() -> None:
    """Shutdown the ZenML server."""
    if logger.isEnabledFor(logging.DEBUG):
        stop_event_loop_lag_monitor()
    run_template_executor().shutdown(wait=True)
    await cleanup_request_manager()


DASHBOARD_REDIRECT_URL = None
if (
    server_config().dashboard_url
    and server_config().deployment_type == ServerDeploymentType.CLOUD
):
    DASHBOARD_REDIRECT_URL = server_config().dashboard_url

if not DASHBOARD_REDIRECT_URL:
    app.mount(
        "/assets",
        StaticFiles(
            directory=relative_path(
                os.path.join(DASHBOARD_DIRECTORY, "assets")
            ),
            check_dir=False,
        ),
    )


# Basic Health Endpoint
@app.head(HEALTH, include_in_schema=False)
@app.get(HEALTH)
async def health() -> str:
    """Get health status of the server.

    Returns:
        String representing the health status of the server.
    """
    return "OK"


# Basic Ready Endpoint
@app.head(READY, include_in_schema=False)
@app.get(READY)
async def ready() -> str:
    """Get readiness status of the server.

    Returns:
        String representing the readiness status of the server.
    """
    return "OK"


templates = Jinja2Templates(directory=relative_path(DASHBOARD_DIRECTORY))


@app.get("/", include_in_schema=False)
async def dashboard(request: Request) -> Any:
    """Dashboard endpoint.

    Args:
        request: Request object.

    Returns:
        The ZenML dashboard.

    Raises:
        HTTPException: If the dashboard files are not included.
    """
    if DASHBOARD_REDIRECT_URL:
        return RedirectResponse(url=DASHBOARD_REDIRECT_URL)

    if not os.path.isfile(
        os.path.join(relative_path(DASHBOARD_DIRECTORY), "index.html")
    ):
        raise HTTPException(status_code=404)
    return templates.TemplateResponse("index.html", {"request": request})


app.include_router(actions_endpoints.router)
app.include_router(artifact_endpoint.artifact_router)
app.include_router(artifact_version_endpoints.artifact_version_router)
app.include_router(auth_endpoints.router)
app.include_router(devices_endpoints.router)
app.include_router(code_repositories_endpoints.router)
app.include_router(plugin_endpoints.plugin_router)
app.include_router(event_source_endpoints.event_source_router)
app.include_router(flavors_endpoints.router)
app.include_router(logs_endpoints.router)
app.include_router(models_endpoints.router)
app.include_router(model_versions_endpoints.router)
app.include_router(model_versions_endpoints.model_version_artifacts_router)
app.include_router(model_versions_endpoints.model_version_pipeline_runs_router)
app.include_router(pipelines_endpoints.router)
app.include_router(pipeline_builds_endpoints.router)
app.include_router(pipeline_deployments_endpoints.router)
app.include_router(runs_endpoints.router)
app.include_router(run_metadata_endpoints.router)
app.include_router(run_templates_endpoints.router)
app.include_router(schedule_endpoints.router)
app.include_router(secrets_endpoints.router)
app.include_router(secrets_endpoints.op_router)
app.include_router(server_endpoints.router)
app.include_router(service_accounts_endpoints.router)
app.include_router(service_connectors_endpoints.router)
app.include_router(service_connectors_endpoints.types_router)
app.include_router(service_endpoints.router)
app.include_router(stack_deployment_endpoints.router)
app.include_router(stacks_endpoints.router)
app.include_router(stack_components_endpoints.router)
app.include_router(stack_components_endpoints.types_router)
app.include_router(steps_endpoints.router)
app.include_router(tags_endpoints.router)
app.include_router(tag_resource_endpoints.router)
app.include_router(triggers_endpoints.router)
app.include_router(users_endpoints.router)
app.include_router(users_endpoints.current_user_router)
app.include_router(webhook_endpoints.router)
app.include_router(projects_endpoints.workspace_router)
app.include_router(projects_endpoints.router)

# When the auth scheme is set to EXTERNAL, users cannot be managed via the
# API.
if server_config().auth_scheme != AuthScheme.EXTERNAL:
    app.include_router(users_endpoints.activation_router)


def get_root_static_files() -> List[str]:
    """Get the list of static files in the root dashboard directory.

    These files are static files that are not in the /static subdirectory
    that need to be served as static files under the root URL path.

    Returns:
        List of static files in the root directory.
    """
    root_path = relative_path(DASHBOARD_DIRECTORY)
    if not os.path.isdir(root_path):
        return []
    files = []
    for file in os.listdir(root_path):
        if file == "index.html":
            # this is served separately
            continue
        if isfile(os.path.join(root_path, file)):
            files.append(file)
    return files


# save these globally to avoid having to poll the filesystem on every request
root_static_files = get_root_static_files()


@app.get(
    API + "/{invalid_api_path:path}", status_code=404, include_in_schema=False
)
async def invalid_api(invalid_api_path: str) -> None:
    """Invalid API endpoint.

    All API endpoints that are not defined in the API routers will be
    redirected to this endpoint and will return a 404 error.

    Args:
        invalid_api_path: Invalid API path.

    Raises:
        HTTPException: 404 error.
    """
    logger.debug(f"Invalid API path requested: {invalid_api_path}")
    raise HTTPException(status_code=404)


@app.get("/{file_path:path}", include_in_schema=False)
async def catch_all(request: Request, file_path: str) -> Any:
    """Dashboard endpoint.

    Args:
        request: Request object.
        file_path: Path to a file in the dashboard root folder.

    Returns:
        The ZenML dashboard.
    """
    if DASHBOARD_REDIRECT_URL:
        return RedirectResponse(url=DASHBOARD_REDIRECT_URL)
    # some static files need to be served directly from the root dashboard
    # directory
    if file_path and file_path in root_static_files:
        logger.debug(f"Returning static file: {file_path}")
        full_path = os.path.join(relative_path(DASHBOARD_DIRECTORY), file_path)
        return FileResponse(full_path)

    # everything else is directed to the index.html file that hosts the
    # single-page application
    return templates.TemplateResponse("index.html", {"request": request})



"""Add server settings [46506f72f0ed].

Revision ID: 46506f72f0ed
Revises: cc9894cb58aa
Create Date: 2024-04-17 14:17:08.142652

"""

from uuid import UUID

import sqlalchemy as sa
import sqlmodel
from alembic import op

from zenml.utils.time_utils import utc_now

# revision identifiers, used by Alembic.
revision = "46506f72f0ed"
down_revision = "cc9894cb58aa"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    from zenml.config.global_config import GlobalConfiguration

    op.create_table(
        "server_settings",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column(
            "server_name", sqlmodel.sql.sqltypes.AutoString(), nullable=False
        ),
        sa.Column(
            "logo_url", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.Column("enable_analytics", sa.Boolean(), nullable=False),
        sa.Column("display_announcements", sa.Boolean(), nullable=False),
        sa.Column("display_updates", sa.Boolean(), nullable=False),
        sa.Column("onboarding_state", sa.TEXT(), nullable=True),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )

    # Get metadata from current connection
    meta = sa.MetaData()

    # Pass in tuple with tables we want to reflect, otherwise whole database
    # will get reflected
    meta.reflect(only=("identity", "server_settings"), bind=op.get_bind())

    # Fetch the deployment id from the identity table
    deployment_id = (
        op.get_bind()
        .execute(sa.select(meta.tables["identity"].c.id).limit(1))
        .scalar_one()
    )

    # Prefill the settings table with a single row that contains the deployment
    # id extracted from the former identity table and some default settings
    op.bulk_insert(
        sa.Table(
            "server_settings",
            meta,
        ),
        [
            {
                "id": deployment_id,
                # Use the deployment ID as the default server name
                "server_name": str(UUID(deployment_id)),
                # Mark existing servers as active
                "active": True,
                # Enable analytics if already enabled through the global config
                "enable_analytics": GlobalConfiguration().analytics_opt_in,
                # Enable announcements and updates by default
                "display_announcements": True,
                "display_updates": True,
                # Set the updated timestamp to the current time
                "updated": utc_now(),
            },
        ],
    )

    op.drop_table("identity")


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    op.create_table(
        "identity",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )

    # Get metadata from current connection
    meta = sa.MetaData()

    # Pass in tuple with tables we want to reflect, otherwise whole database
    # will get reflected
    meta.reflect(only=("identity", "server_settings"), bind=op.get_bind())

    # Fetch the deployment id from the settings table
    deployment_id = (
        op.get_bind()
        .execute(sa.select(meta.tables["server_settings"].c.id).limit(1))
        .scalar_one()
    )

    # Prefill the identity table with a single row that contains the deployment
    # id extracted from the settings table
    op.bulk_insert(
        sa.Table(
            "identity",
            meta,
        ),
        [
            {
                "id": deployment_id,
            },
        ],
    )

    op.drop_table("server_settings")



"""add unique name constraints [1f9d1cd00b90].

Revision ID: 1f9d1cd00b90
Revises: f76a368a25a5
Create Date: 2025-02-22 20:18:34.258987

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session

from zenml.logger import get_logger

logger = get_logger(__name__)

# revision identifiers, used by Alembic.
revision = "1f9d1cd00b90"
down_revision = "f76a368a25a5"
branch_labels = None
depends_on = None


def resolve_duplicate_names(
    table_name: str, other_columns: list[str], session: Session
) -> None:
    """Resolve duplicate entities.

    Args:
        table_name: The name of the table to resolve duplicate entities for.
        other_columns: The columns that are part of the unique constraint,
            excluding the name column.
        session: The SQLAlchemy session to use.
    """
    columns = ["name"] + other_columns
    duplicates = session.execute(
        sa.text(
            f"""
                SELECT id, name
                FROM `{table_name}`
                WHERE ({", ".join(columns)}) IN (
                    SELECT {", ".join(columns)}
                    FROM `{table_name}`
                    GROUP BY {", ".join(columns)}
                    HAVING COUNT(*) > 1
                )
            """  # nosec B608
        )
    )
    for id_, name in list(duplicates)[1:]:
        logger.warning(f"Duplicate {table_name}: {name} (id: {id_})")
        session.execute(
            sa.text(
                f"""
                    UPDATE {table_name}
                    SET name = :new_name
                    WHERE id = :id_
                """  # nosec B608
            ),
            params={"id_": id_, "new_name": f"{name}_{id_[:6]}"},
        )


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    bind = op.get_bind()
    session = Session(bind=bind)

    resolve_duplicate_names("action", ["workspace_id"], session)

    with op.batch_alter_table("action", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_action_name_in_workspace", ["name", "workspace_id"]
        )

    resolve_duplicate_names("api_key", ["service_account_id"], session)

    with op.batch_alter_table("api_key", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_api_key_name_in_service_account",
            ["name", "service_account_id"],
        )

    resolve_duplicate_names("artifact", ["workspace_id"], session)

    with op.batch_alter_table("artifact", schema=None) as batch_op:
        batch_op.drop_constraint("unique_artifact_name", type_="unique")
        batch_op.create_unique_constraint(
            "unique_artifact_name_in_workspace", ["name", "workspace_id"]
        )

    resolve_duplicate_names("code_repository", ["workspace_id"], session)

    with op.batch_alter_table("code_repository", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_code_repository_name_in_workspace",
            ["name", "workspace_id"],
        )

    resolve_duplicate_names("event_source", ["workspace_id"], session)

    with op.batch_alter_table("event_source", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_event_source_name_in_workspace", ["name", "workspace_id"]
        )

    resolve_duplicate_names("flavor", ["type"], session)

    with op.batch_alter_table("flavor", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_flavor_name_and_type", ["name", "type"]
        )

    resolve_duplicate_names("schedule", ["workspace_id"], session)

    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_schedule_name_in_workspace", ["name", "workspace_id"]
        )

    resolve_duplicate_names("secret", ["private", "user_id"], session)

    with op.batch_alter_table("secret", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_secret_name_private_scope_user",
            ["name", "private", "user_id"],
        )

    resolve_duplicate_names("service_connector", [], session)

    with op.batch_alter_table("service_connector", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_service_connector_name", ["name"]
        )

    resolve_duplicate_names("stack", [], session)

    with op.batch_alter_table("stack", schema=None) as batch_op:
        batch_op.create_unique_constraint("unique_stack_name", ["name"])

    resolve_duplicate_names("stack_component", ["type"], session)

    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_component_name_and_type", ["name", "type"]
        )

    resolve_duplicate_names("tag", [], session)

    with op.batch_alter_table("tag", schema=None) as batch_op:
        batch_op.create_unique_constraint("unique_tag_name", ["name"])

    resolve_duplicate_names("trigger", ["workspace_id"], session)

    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_trigger_name_in_workspace", ["name", "workspace_id"]
        )

    resolve_duplicate_names("workspace", [], session)

    with op.batch_alter_table("workspace", schema=None) as batch_op:
        batch_op.create_unique_constraint("unique_workspace_name", ["name"])


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("workspace", schema=None) as batch_op:
        batch_op.drop_constraint("unique_workspace_name", type_="unique")

    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_trigger_name_in_workspace", type_="unique"
        )

    with op.batch_alter_table("tag", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_tag_name_in_workspace", type_="unique"
        )

    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_component_name_and_type", type_="unique"
        )

    with op.batch_alter_table("stack", schema=None) as batch_op:
        batch_op.drop_constraint("unique_stack_name", type_="unique")

    with op.batch_alter_table("service_connector", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_service_connector_name", type_="unique"
        )

    with op.batch_alter_table("secret", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_secret_name_and_private_scope", type_="unique"
        )

    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_schedule_name_in_workspace", type_="unique"
        )

    with op.batch_alter_table("flavor", schema=None) as batch_op:
        batch_op.drop_constraint("unique_flavor_name_and_type", type_="unique")

    with op.batch_alter_table("event_source", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_event_source_name_in_workspace", type_="unique"
        )

    with op.batch_alter_table("code_repository", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_code_repository_name_in_workspace", type_="unique"
        )

    with op.batch_alter_table("artifact", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_artifact_name_in_workspace", type_="unique"
        )
        batch_op.create_unique_constraint("unique_artifact_name", ["name"])

    with op.batch_alter_table("api_key", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_api_key_name_in_service_account", type_="unique"
        )

    with op.batch_alter_table("action", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_action_name_in_workspace", type_="unique"
        )



"""Release [0.31.0].

Revision ID: 0.31.0
Revises: 248dfd320b68
Create Date: 2022-12-23 16:05:04.535949

"""

# revision identifiers, used by Alembic.
revision = "0.31.0"
down_revision = "248dfd320b68"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Modal integration for cloud-native step execution.

The Modal integration sub-module provides a step operator flavor that allows
executing steps on Modal's cloud infrastructure.
"""
from typing import List, Type

from zenml.integrations.constants import MODAL
from zenml.integrations.integration import Integration
from zenml.stack import Flavor

MODAL_STEP_OPERATOR_FLAVOR = "modal"


class ModalIntegration(Integration):
    """Definition of Modal integration for ZenML."""

    NAME = MODAL
    REQUIREMENTS = ["modal>=0.64.49,<1"]

    @classmethod
    def flavors(cls) -> List[Type[Flavor]]:
        """Declare the stack component flavors for the Modal integration.

        Returns:
            List of new stack component flavors.
        """
        from zenml.integrations.modal.flavors import ModalStepOperatorFlavor

        return [ModalStepOperatorFlavor]





#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""HashiCorp Vault Secrets Store implementation."""

from typing import (
    ClassVar,
    Dict,
    Optional,
    Type,
)
from uuid import UUID

import hvac  # type: ignore[import-untyped]
from hvac.exceptions import (  # type: ignore[import-untyped]
    InvalidPath,
    VaultError,
)
from pydantic import ConfigDict

from zenml.config.secrets_store_config import SecretsStoreConfiguration
from zenml.enums import (
    SecretsStoreType,
)
from zenml.logger import get_logger
from zenml.utils.secret_utils import PlainSerializedSecretStr
from zenml.zen_stores.secrets_stores.base_secrets_store import (
    BaseSecretsStore,
)

logger = get_logger(__name__)


HVAC_ZENML_SECRET_NAME_PREFIX = "zenml"
ZENML_VAULT_SECRET_VALUES_KEY = "zenml_secret_values"
ZENML_VAULT_SECRET_METADATA_KEY = "zenml_secret_metadata"


class HashiCorpVaultSecretsStoreConfiguration(SecretsStoreConfiguration):
    """HashiCorp Vault secrets store configuration.

    Attributes:
        type: The type of the store.
        vault_addr: The url of the Vault server. If not set, the value will be
            loaded from the VAULT_ADDR environment variable, if configured.
        vault_token: The token used to authenticate with the Vault server. If
            not set, the token will be loaded from the VAULT_TOKEN environment
            variable or from the ~/.vault-token file, if configured.
        vault_namespace: The Vault Enterprise namespace.
        mount_point: The mount point to use for all secrets.
        max_versions: The maximum number of secret versions to keep.
    """

    type: SecretsStoreType = SecretsStoreType.HASHICORP

    vault_addr: str
    vault_token: Optional[PlainSerializedSecretStr] = None
    vault_namespace: Optional[str] = None
    mount_point: Optional[str] = None
    max_versions: int = 1
    model_config = ConfigDict(extra="ignore")


class HashiCorpVaultSecretsStore(BaseSecretsStore):
    """Secrets store implementation that uses the HashiCorp Vault API.

    This secrets store implementation uses the HashiCorp Vault API to
    store secrets. It allows a single HashiCorp Vault server to be shared with
    other ZenML deployments as well as other third party users and applications.

    Here are some implementation highlights:

    * the name/ID of an HashiCorp Vault secret is derived from the ZenML secret
    UUID and a `zenml` prefix in the form `zenml/{zenml_secret_uuid}`. This
    clearly identifies a secret as being managed by ZenML in the HashiCorp Vault
    server.

    * given that HashiCorp Vault secrets do not support attaching arbitrary
    metadata in the form of label or tags, we store the entire ZenML secret
    metadata alongside the secret values in the HashiCorp Vault secret value.

    Attributes:
        config: The configuration of the HashiCorp Vault secrets store.
        TYPE: The type of the store.
        CONFIG_TYPE: The type of the store configuration.
    """

    config: HashiCorpVaultSecretsStoreConfiguration
    TYPE: ClassVar[SecretsStoreType] = SecretsStoreType.HASHICORP
    CONFIG_TYPE: ClassVar[Type[SecretsStoreConfiguration]] = (
        HashiCorpVaultSecretsStoreConfiguration
    )

    _client: Optional[hvac.Client] = None

    @property
    def client(self) -> hvac.Client:
        """Initialize and return the HashiCorp Vault client.

        Returns:
            The HashiCorp Vault client.
        """
        if self._client is None:
            # Initialize the HashiCorp Vault client with the
            # credentials from the configuration.
            self._client = hvac.Client(
                url=self.config.vault_addr,
                token=self.config.vault_token.get_secret_value()
                if self.config.vault_token
                else None,
                namespace=self.config.vault_namespace,
            )
            self._client.secrets.kv.v2.configure(
                max_versions=self.config.max_versions,
            )
            if self.config.mount_point:
                self._client.secrets.kv.v2.configure(
                    mount_point=self.config.mount_point,
                )
        return self._client

    # ====================================
    # Secrets Store interface implementation
    # ====================================

    # --------------------------------
    # Initialization and configuration
    # --------------------------------

    def _initialize(self) -> None:
        """Initialize the HashiCorp Vault secrets store."""
        logger.debug("Initializing HashiCorpVaultSecretsStore")

        # Initialize the HashiCorp Vault client early, just to catch any
        # configuration or authentication errors early, before the Secrets
        # Store is used.
        _ = self.client

    # ------
    # Secrets
    # ------

    @staticmethod
    def _get_vault_secret_id(
        secret_id: UUID,
    ) -> str:
        """Get the HashiCorp Vault secret ID corresponding to a ZenML secret ID.

        The convention used for HashiCorp Vault secret names is to use the ZenML
        secret UUID prefixed with `zenml` as the HashiCorp Vault secret name,
        i.e. `zenml/<secret_uuid>`.

        Args:
            secret_id: The ZenML secret ID.

        Returns:
            The HashiCorp Vault secret name.
        """
        return f"{HVAC_ZENML_SECRET_NAME_PREFIX}/{str(secret_id)}"

    def store_secret_values(
        self,
        secret_id: UUID,
        secret_values: Dict[str, str],
    ) -> None:
        """Store secret values for a new secret.

        Args:
            secret_id: ID of the secret.
            secret_values: Values for the secret.

        Raises:
            RuntimeError: If the HashiCorp Vault API returns an unexpected
                error.
        """
        vault_secret_id = self._get_vault_secret_id(secret_id)

        metadata = self._get_secret_metadata(secret_id=secret_id)

        try:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=vault_secret_id,
                # Store the ZenML secret metadata alongside the secret values
                secret={
                    ZENML_VAULT_SECRET_VALUES_KEY: secret_values,
                    ZENML_VAULT_SECRET_METADATA_KEY: metadata,
                },
                # Do not allow overwriting an existing secret
                cas=0,
            )
        except VaultError as e:
            raise RuntimeError(f"Error creating secret: {e}")

        logger.debug(f"Created HashiCorp Vault secret: {vault_secret_id}")

    def get_secret_values(self, secret_id: UUID) -> Dict[str, str]:
        """Get the secret values for an existing secret.

        Args:
            secret_id: ID of the secret.

        Returns:
            The secret values.

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
            RuntimeError: If the HashiCorp Vault API returns an unexpected
                error.
        """
        vault_secret_id = self._get_vault_secret_id(secret_id)

        try:
            vault_secret = (
                self.client.secrets.kv.v2.read_secret(
                    path=vault_secret_id,
                )
                .get("data", {})
                .get("data", {})
            )
        except InvalidPath as e:
            raise KeyError(
                f"Can't find the secret values for secret ID '{secret_id}' "
                f"in the secrets store back-end: {str(e)}"
            ) from e
        except VaultError as e:
            raise RuntimeError(
                f"Error fetching secret with ID {secret_id} {e}"
            )

        try:
            metadata = vault_secret[ZENML_VAULT_SECRET_METADATA_KEY]
            values = vault_secret[ZENML_VAULT_SECRET_VALUES_KEY]
        except (KeyError, ValueError) as e:
            raise KeyError(
                f"Secret could not be retrieved: missing required metadata: {e}"
            )

        if not isinstance(values, dict) or not isinstance(metadata, dict):
            raise RuntimeError(
                f"HashiCorp Vault secret values for secret {vault_secret_id} "
                "could not be retrieved: invalid type for metadata or values"
            )

        # The _verify_secret_metadata method raises a KeyError if the
        # secret is not valid or does not belong to this server. Here we
        # simply pass the exception up the stack, as if the secret was not found
        # in the first place.
        self._verify_secret_metadata(
            secret_id=secret_id,
            metadata=metadata,
        )

        logger.debug(f"Fetched HashiCorp Vault secret: {vault_secret_id}")

        return values

    def update_secret_values(
        self,
        secret_id: UUID,
        secret_values: Dict[str, str],
    ) -> None:
        """Updates secret values for an existing secret.

        Args:
            secret_id: The ID of the secret to be updated.
            secret_values: The new secret values.

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
            RuntimeError: If the HashiCorp Vault API returns an unexpected
                error.
        """
        vault_secret_id = self._get_vault_secret_id(secret_id)

        # Convert the ZenML secret metadata to HashiCorp Vault tags
        metadata = self._get_secret_metadata(secret_id=secret_id)

        try:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=vault_secret_id,
                # Store the ZenML secret metadata alongside the secret values
                secret={
                    ZENML_VAULT_SECRET_VALUES_KEY: secret_values,
                    ZENML_VAULT_SECRET_METADATA_KEY: metadata,
                },
            )
        except InvalidPath:
            raise KeyError(f"Secret with ID {secret_id} does not exist.")
        except VaultError as e:
            raise RuntimeError(f"Error updating secret {secret_id}: {e}")

        logger.debug(f"Updated HashiCorp Vault secret: {vault_secret_id}")

    def delete_secret_values(self, secret_id: UUID) -> None:
        """Deletes secret values for an existing secret.

        Args:
            secret_id: The ID of the secret.

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
            RuntimeError: If the HashiCorp Vault API returns an unexpected
                error.
        """
        vault_secret_id = self._get_vault_secret_id(secret_id)

        try:
            self.client.secrets.kv.v2.delete_metadata_and_all_versions(
                path=vault_secret_id,
            )
        except InvalidPath:
            raise KeyError(f"Secret with ID {secret_id} does not exist.")
        except VaultError as e:
            raise RuntimeError(
                f"Error deleting secret with ID {secret_id}: {e}"
            )

        logger.debug(f"Deleted HashiCorp Vault secret: {vault_secret_id}")



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Feast feature store flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.feature_stores.base_feature_store import (
    BaseFeatureStoreConfig,
    BaseFeatureStoreFlavor,
)
from zenml.integrations.feast import FEAST_FEATURE_STORE_FLAVOR

if TYPE_CHECKING:
    from zenml.integrations.feast.feature_stores import FeastFeatureStore


class FeastFeatureStoreConfig(BaseFeatureStoreConfig):
    """Config for Feast feature store.

    Configuration for connecting to Feast feature stores.
    Field descriptions are defined inline using Field() descriptors.
    """

    online_host: str = Field(
        default="localhost",
        description="Online feature store host address (typically Redis server).",
    )
    online_port: int = Field(
        default=6379, description="Online feature store port number."
    )
    feast_repo: str = Field(
        description="Local filesystem path to the Feast repository with feature definitions."
    )

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return (
            self.online_host == "localhost" or self.online_host == "127.0.0.1"
        )


class FeastFeatureStoreFlavor(BaseFeatureStoreFlavor):
    """Feast Feature store flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return FEAST_FEATURE_STORE_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/feature_store/feast.png"

    @property
    def config_class(self) -> Type[FeastFeatureStoreConfig]:
        """Returns FeastFeatureStoreConfig config class.

        Returns:
                The config class.
        """
        """Config class for this flavor."""
        return FeastFeatureStoreConfig

    @property
    def implementation_class(self) -> Type["FeastFeatureStore"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.feast.feature_stores import FeastFeatureStore

        return FeastFeatureStore
